from __future__ import annotations

import re
from functools import wraps
from itertools import zip_longest
from typing import Callable, Union, overload

from semantic_release.const import SEMVER_REGEX
from semantic_release.enums import LevelBump
from semantic_release.errors import InvalidVersion
from semantic_release.globals import logger
from semantic_release.helpers import check_tag_format

# Very heavily inspired by semver.version:_comparator, I don't think there's
# a cleaner way to do this
# https://github.com/python-semver/python-semver/blob/b5317af9a7e99e6a86df98320e73be72d5adf0de/src/semver/version.py#L32
VersionComparable = Union["Version", str]
VersionComparator = Callable[["Version", "Version"], bool]


@overload
def _comparator(
    *,
    type_guard: bool,
) -> Callable[[VersionComparator], VersionComparator]: ...


@overload
def _comparator(
    method: VersionComparator, *, type_guard: bool = True
) -> VersionComparator: ...


def _comparator(
    method: VersionComparator | None = None, *, type_guard: bool = True
) -> VersionComparator | Callable[[VersionComparator], VersionComparator]:
    """
    wrap a `Version` binop method to guard types and try to parse strings into Versions.
    use `type_guard = False` for `__eq__` and `__neq__` to make them return False if the
    wrong type is used, instead of erroring.
    """
    if method is None:
        return lambda method: _comparator(method, type_guard=type_guard)

    @wraps(method)
    def _wrapper(self: Version, other: VersionComparable) -> bool:
        if not isinstance(other, (str, Version)):
            return False if not type_guard else NotImplemented
        if isinstance(other, str):
            try:
                other_v = self.parse(
                    other,
                    tag_format=self.tag_format,
                    prerelease_token=self.prerelease_token,
                )
            except InvalidVersion as ex:
                raise TypeError(str(ex)) from ex
        else:
            other_v = other

        return method(self, other_v)  # type: ignore[misc]

    return _wrapper


class Version:
    _VERSION_REGEX = SEMVER_REGEX

    def __init__(
        self,
        major: int,
        minor: int,
        patch: int,
        *,
        prerelease_token: str = "rc",  # noqa: S107
        prerelease_revision: int | None = None,
        build_metadata: str = "",
        tag_format: str = "v{version}",
    ) -> None:
        self.major = major
        self.minor = minor
        self.patch = patch
        self.prerelease_token = prerelease_token
        self.prerelease_revision = prerelease_revision
        self.build_metadata = build_metadata
        self._tag_format = tag_format

    @property
    def tag_format(self) -> str:
        return self._tag_format

    @tag_format.setter
    def tag_format(self, new_format: str) -> None:
        check_tag_format(new_format)
        self._tag_format = new_format

    # Maybe cache?
    @classmethod
    def parse(
        cls,
        version_str: str,
        tag_format: str = "v{version}",
        prerelease_token: str = "rc",  # noqa: S107
    ) -> Version:
        """
        Parse version string to a Version instance.
        Inspired by `semver.version:VersionInfo.parse`, this implementation doesn't
        allow optional minor and patch versions.

        :param prerelease_token: will be ignored if the version string is a prerelease,
            the parsed token from `version_str` will be used instead.
        """
        if not isinstance(version_str, str):
            raise InvalidVersion(f"{version_str!r} cannot be parsed as a Version")

        logger.debug("attempting to parse string %r as Version", version_str)
        match = cls._VERSION_REGEX.fullmatch(version_str)
        if not match:
            raise InvalidVersion(f"{version_str!r} is not a valid Version")

        prerelease = match.group("prerelease")
        if prerelease:
            pm = re.match(r"(?P<token>[a-zA-Z0-9-\.]+)\.(?P<revision>\d+)", prerelease)
            if not pm:
                raise NotImplementedError(
                    f"{cls.__qualname__} currently supports only prereleases "
                    r"of the format (-([a-zA-Z0-9-])\.\(\d+)), for example "
                    r"'1.2.3-my-custom-3rc.4'."
                )
            prerelease_token, prerelease_revision = pm.groups()
            logger.debug(
                "parsed prerelease_token %s, prerelease_revision %s from version "
                "string %s",
                prerelease_token,
                prerelease_revision,
                version_str,
            )
        else:
            prerelease_revision = None
            logger.debug("version string %s parsed as a non-prerelease", version_str)

        build_metadata = match.group("buildmetadata") or ""
        logger.debug(
            "parsed build metadata %r from version string %s",
            build_metadata,
            version_str,
        )

        return Version(
            int(match.group("major")),
            int(match.group("minor")),
            int(match.group("patch")),
            prerelease_token=prerelease_token,
            prerelease_revision=(
                int(prerelease_revision) if prerelease_revision else None
            ),
            build_metadata=build_metadata,
            tag_format=tag_format,
        )

    @property
    def is_prerelease(self) -> bool:
        return self.prerelease_revision is not None

    def __str__(self) -> str:
        full = f"{self.major}.{self.minor}.{self.patch}"
        prerelease = (
            f"-{self.prerelease_token}.{self.prerelease_revision}"
            if self.prerelease_revision
            else ""
        )
        build_metadata = f"+{self.build_metadata}" if self.build_metadata else ""
        return f"{full}{prerelease}{build_metadata}"

    def __repr__(self) -> str:
        prerelease_token_repr = (
            repr(self.prerelease_token) if self.prerelease_token is not None else None
        )
        prerelease_revision_repr = (
            repr(self.prerelease_revision)
            if self.prerelease_revision is not None
            else None
        )
        build_metadata_repr = (
            repr(self.build_metadata) if self.build_metadata is not None else None
        )
        return (
            f"{type(self).__qualname__}("
            + ", ".join(
                (
                    f"major={self.major}",
                    f"minor={self.minor}",
                    f"patch={self.patch}",
                    f"prerelease_token={prerelease_token_repr}",
                    f"prerelease_revision={prerelease_revision_repr}",
                    f"build_metadata={build_metadata_repr}",
                    f"tag_format={self.tag_format!r}",
                )
            )
            + ")"
        )

    def as_tag(self) -> str:
        return self.tag_format.format(version=str(self))

    def as_major_tag(self) -> str:
        return self.tag_format.format(version=f"{self.major}")

    def as_minor_tag(self) -> str:
        return self.tag_format.format(version=f"{self.major}.{self.minor}")

    def as_patch_tag(self) -> str:
        return self.tag_format.format(version=f"{self.major}.{self.minor}.{self.patch}")

    def as_semver_tag(self) -> str:
        return f"v{self!s}"

    def bump(self, level: LevelBump) -> Version:
        """
        Return a new Version instance according to the level specified to bump.
        Note this will intentionally drop the build metadata - that should be added
        elsewhere for the specific build producing this version.
        """
        if type(level) != LevelBump:
            raise TypeError(f"Unexpected level {level!r}: expected {LevelBump!r}")

        logger.debug("performing a %s level bump", level)
        if level is LevelBump.MAJOR:
            return Version(
                self.major + 1,
                0,
                0,
                prerelease_token=self.prerelease_token,
                prerelease_revision=1 if self.is_prerelease else None,
                tag_format=self.tag_format,
            )
        if level is LevelBump.MINOR:
            return Version(
                self.major,
                self.minor + 1,
                0,
                prerelease_token=self.prerelease_token,
                prerelease_revision=1 if self.is_prerelease else None,
                tag_format=self.tag_format,
            )
        if level is LevelBump.PATCH:
            return Version(
                self.major,
                self.minor,
                self.patch + 1,
                prerelease_token=self.prerelease_token,
                prerelease_revision=1 if self.is_prerelease else None,
                tag_format=self.tag_format,
            )
        if level is LevelBump.PRERELEASE_REVISION:
            return Version(
                self.major,
                self.minor,
                self.patch,
                prerelease_token=self.prerelease_token,
                prerelease_revision=1
                if not self.is_prerelease
                else (self.prerelease_revision or 0) + 1,
                tag_format=self.tag_format,
            )
        # for consistency, this creates a new instance regardless
        # only other option is level is LevelBump.NO_RELEASE
        return Version(
            self.major,
            self.minor,
            self.patch,
            prerelease_token=self.prerelease_token,
            prerelease_revision=self.prerelease_revision,
            tag_format=self.tag_format,
        )

    # Enables Version + LevelBump.<level>
    __add__ = bump

    def __hash__(self) -> int:
        # If we use str(self) we don't capture tag_format, so another
        # instance with a tag_format "special_{version}_format" would
        # collide with an instance using "v{version}"/other format
        return hash(self.__repr__())

    @_comparator(type_guard=False)
    def __eq__(self, other: Version) -> bool:  # type: ignore[override]
        # https://semver.org/#spec-item-11 -
        # build metadata is not used for comparison
        return all(
            getattr(self, attr) == getattr(other, attr)
            for attr in (
                "major",
                "minor",
                "patch",
                "prerelease_token",
                "prerelease_revision",
            )
        )

    @_comparator(type_guard=False)
    def __neq__(self, other: Version) -> bool:
        return not self.__eq__(other)

    # mypy wants to compare signature types with __lt__,
    # but can't because of the decorator
    @_comparator
    def __gt__(self, other: Version) -> bool:  # type: ignore[has-type]
        # https://semver.org/#spec-item-11 -
        # build metadata is not used for comparison

        # Note we only support the following versioning currently, which
        # is a subset of the full spec:
        # (\d+\.\d+\.\d+)(-\w+\.\d+)?(\+.*)?
        if self.major != other.major:
            return self.major > other.major
        if self.minor != other.minor:
            return self.minor > other.minor
        if self.patch != other.patch:
            return self.patch > other.patch
        # If just one is a prerelease, then self > other if other is the prerelease
        # If neither are prereleases then they're equal (so return False)
        if not (self.is_prerelease and other.is_prerelease):
            return other.is_prerelease
        # If both are prereleases...
        # According to the semver spec 11.4 there are many other rules for
        # comparing precedence of pre-release versions. Here we just compare
        # the prerelease tokens, and their revision numbers
        if self.prerelease_token != other.prerelease_token:
            for self_tk, other_tk in zip_longest(
                self.prerelease_token.split("."),
                other.prerelease_token.split("."),
                fillvalue=None,
            ):
                if self_tk == other_tk:
                    continue
                if (self_tk is None) ^ (other_tk is None):
                    # Longest token (i.e. non-None) is greater
                    return other_tk is None
                # Lexical sort, e.g. "rc" > "beta" > "alpha"
                # we have eliminated that one or both might be None above,
                # but mypy doesn't recognise this
                return self_tk > other_tk  # type: ignore[operator]
        # We have eliminated that one or both aren't prereleases by the above
        return self.prerelease_revision > other.prerelease_revision  # type: ignore[operator]  # noqa: E501

    # mypy wants to compare signature types with __le__,
    # but can't because of the decorator
    @_comparator
    def __ge__(self, other: Version) -> bool:  # type: ignore[has-type]
        return self.__gt__(other) or self.__eq__(other)

    @_comparator
    def __lt__(self, other: Version) -> bool:
        return not (self.__gt__(other) or self.__eq__(other))

    @_comparator
    def __le__(self, other: Version) -> bool:
        return not self.__gt__(other)

    def __sub__(self, other: Version) -> LevelBump:
        if not isinstance(other, Version):
            return NotImplemented

        if self.major != other.major:
            return LevelBump.MAJOR
        if self.minor != other.minor:
            return LevelBump.MINOR
        if self.patch != other.patch:
            return LevelBump.PATCH
        if self.is_prerelease ^ other.is_prerelease:
            return max(
                self.finalize_version() - other.finalize_version(),
                LevelBump.PRERELEASE_REVISION,
            )
        if self.prerelease_revision != other.prerelease_revision:
            return LevelBump.PRERELEASE_REVISION
        return LevelBump.NO_RELEASE

    def to_prerelease(
        self, token: str | None = None, revision: int | None = None
    ) -> Version:
        return Version(
            self.major,
            self.minor,
            self.patch,
            prerelease_token=token or self.prerelease_token,
            prerelease_revision=(revision or self.prerelease_revision) or 1,
            tag_format=self.tag_format,
        )

    def finalize_version(self) -> Version:
        return Version(
            self.major,
            self.minor,
            self.patch,
            prerelease_token=self.prerelease_token,
            tag_format=self.tag_format,
        )



from __future__ import annotations

import os
import shutil
from pathlib import Path
from re import MULTILINE, compile as regexp
from typing import TYPE_CHECKING

import pytest
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.changelog.context import ChangelogMode
from semantic_release.cli.const import JINJA2_EXTENSION

from tests.const import CHANGELOG_SUBCMD, MAIN_PROG_NAME
from tests.fixtures.example_project import (
    default_changelog_md_template,
    default_changelog_rst_template,
    default_md_changelog_insertion_flag,
    default_rst_changelog_insertion_flag,
    example_changelog_md,
    example_changelog_rst,
)
from tests.fixtures.repos.git_flow import (
    repo_w_git_flow_conventional_commits,
    repo_w_git_flow_scipy_commits,
)
from tests.util import assert_successful_exit_code

if TYPE_CHECKING:
    from tests.conftest import RunCliFn
    from tests.fixtures.example_project import UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import BuiltRepoResult


@pytest.mark.parametrize(
    "changelog_file, insertion_flag, default_changelog_template, changes_tpl_file",
    [
        (
            # ChangelogOutputFormat.MARKDOWN
            lazy_fixture(example_changelog_md.__name__),
            lazy_fixture(default_md_changelog_insertion_flag.__name__),
            lazy_fixture(default_changelog_md_template.__name__),
            Path(".components", "changes.md.j2"),
        ),
        (
            # ChangelogOutputFormat.RESTRUCTURED_TEXT
            lazy_fixture(example_changelog_rst.__name__),
            lazy_fixture(default_rst_changelog_insertion_flag.__name__),
            lazy_fixture(default_changelog_rst_template.__name__),
            Path(".components", "changes.rst.j2"),
        ),
    ],
)
@pytest.mark.parametrize(
    "repo_result",
    [
        pytest.param(
            lazy_fixture(repo_fixture_name),
            marks=pytest.mark.comprehensive,
        )
        for repo_fixture_name in [
            repo_w_git_flow_conventional_commits.__name__,
            repo_w_git_flow_scipy_commits.__name__,
        ]
    ],
)
def test_changelog_parsing_ignore_merge_commits(
    run_cli: RunCliFn,
    repo_result: BuiltRepoResult,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    example_project_template_dir: Path,
    changelog_file: Path,
    insertion_flag: str,
    default_changelog_template: Path,
    changes_tpl_file: Path,
):
    repo = repo_result["repo"]
    expected_changelog_content = changelog_file.read_text()

    update_pyproject_toml(
        "tool.semantic_release.commit_parser_options.ignore_merge_commits", True
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.insertion_flag",
        insertion_flag,
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.template_dir",
        str(example_project_template_dir.relative_to(repo.working_dir)),
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.exclude_commit_patterns",
        [
            r"""Initial Commit.*""",
        ],
    )

    # Force custom changelog to be a copy of the default changelog
    shutil.copytree(
        src=default_changelog_template.parent,
        dst=example_project_template_dir,
        dirs_exist_ok=True,
    )

    # Remove the "unknown" filter from the changelog template to enable Merge commits
    patch = regexp(
        r'^(#}{% *for type_, commits in commit_objects) if type_ != "unknown"',
        MULTILINE,
    )
    changes_file = example_project_template_dir.joinpath(changes_tpl_file)
    changes_file.write_text(patch.sub(r"\1", changes_file.read_text()))

    # Make sure the prev_changelog_file is the same as the current changelog
    changelog_tpl_file = example_project_template_dir.joinpath(
        changelog_file.name
    ).with_suffix(str.join("", [changelog_file.suffix, JINJA2_EXTENSION]))
    changelog_tpl_file.write_text(
        regexp(r"= ctx.prev_changelog_file").sub(
            rf'= "{changelog_file.name}"', changelog_tpl_file.read_text()
        )
    )

    # Remove the changelog to force re-generation with new configurations
    os.remove(str(changelog_file.resolve()))

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert expected_changelog_content == changelog_file.read_text()



"""Commit parser which looks for emojis to determine the type of commit"""

from __future__ import annotations

import re
from functools import reduce
from itertools import zip_longest
from re import compile as regexp
from textwrap import dedent
from typing import Tuple

from git.objects.commit import Commit
from pydantic.dataclasses import dataclass

from semantic_release.commit_parser._base import CommitParser, ParserOptions
from semantic_release.commit_parser.token import (
    ParsedCommit,
    ParsedMessageResult,
    ParseError,
    ParseResult,
)
from semantic_release.commit_parser.util import (
    deep_copy_commit,
    force_str,
    parse_paragraphs,
)
from semantic_release.enums import LevelBump
from semantic_release.errors import InvalidParserOptions
from semantic_release.globals import logger
from semantic_release.helpers import sort_numerically, text_reducer


@dataclass
class EmojiParserOptions(ParserOptions):
    """Options dataclass for EmojiCommitParser"""

    major_tags: Tuple[str, ...] = (":boom:",)
    """Commit-type prefixes that should result in a major release bump."""

    minor_tags: Tuple[str, ...] = (
        ":sparkles:",
        ":children_crossing:",
        ":lipstick:",
        ":iphone:",
        ":egg:",
        ":chart_with_upwards_trend:",
    )
    """Commit-type prefixes that should result in a minor release bump."""

    patch_tags: Tuple[str, ...] = (
        ":ambulance:",
        ":lock:",
        ":bug:",
        ":zap:",
        ":goal_net:",
        ":alien:",
        ":wheelchair:",
        ":speech_balloon:",
        ":mag:",
        ":apple:",
        ":penguin:",
        ":checkered_flag:",
        ":robot:",
        ":green_apple:",
    )
    """Commit-type prefixes that should result in a patch release bump."""

    other_allowed_tags: Tuple[str, ...] = (":memo:", ":checkmark:")
    """Commit-type prefixes that are allowed but do not result in a version bump."""

    allowed_tags: Tuple[str, ...] = (
        *major_tags,
        *minor_tags,
        *patch_tags,
        *other_allowed_tags,
    )
    """All commit-type prefixes that are allowed."""

    default_bump_level: LevelBump = LevelBump.NO_RELEASE
    """The minimum bump level to apply to valid commit message."""

    parse_linked_issues: bool = False
    """
    Whether to parse linked issues from the commit message.

    Issue identification is not defined in the Gitmoji specification, so this parser
    will not attempt to parse issues by default. If enabled, the parser will use the
    same identification as GitHub, GitLab, and BitBucket use for linking issues, which
    is to look for a git commit message footer starting with "Closes:", "Fixes:",
    or "Resolves:" then a space, and then the issue identifier. The line prefix
    can be singular or plural and it is not case-sensitive but must have a colon and
    a whitespace separator.
    """

    parse_squash_commits: bool = True
    """Toggle flag for whether or not to parse squash commits"""

    ignore_merge_commits: bool = True
    """Toggle flag for whether or not to ignore merge commits"""

    @property
    def tag_to_level(self) -> dict[str, LevelBump]:
        """A mapping of commit tags to the level bump they should result in."""
        return self._tag_to_level

    def __post_init__(self) -> None:
        self._tag_to_level: dict[str, LevelBump] = {
            str(tag): level
            for tag, level in [
                # we have to do a type ignore as zip_longest provides a type that is not specific enough
                # for our expected output. Due to the empty second array, we know the first is always longest
                # and that means no values in the first entry of the tuples will ever be a LevelBump. We
                # apply a str() to make mypy happy although it will never happen.
                *zip_longest(self.allowed_tags, (), fillvalue=self.default_bump_level),
                *zip_longest(self.patch_tags, (), fillvalue=LevelBump.PATCH),
                *zip_longest(self.minor_tags, (), fillvalue=LevelBump.MINOR),
                *zip_longest(self.major_tags, (), fillvalue=LevelBump.MAJOR),
            ]
            if "|" not in str(tag)
        }


class EmojiCommitParser(CommitParser[ParseResult, EmojiParserOptions]):
    """
    Parse a commit using an emoji in the subject line.
    When multiple emojis are encountered, the one with the highest bump
    level is used. If there are multiple emojis on the same level, the
    we use the one listed earliest in the configuration.
    If the message does not contain any known emojis, then the level to bump
    will be 0 and the type of change "Other". This parser never raises
    UnknownCommitMessageStyleError.
    Emojis are not removed from the description, and will appear alongside
    the commit subject in the changelog.
    """

    # TODO: Deprecate in lieu of get_default_options()
    parser_options = EmojiParserOptions

    def __init__(self, options: EmojiParserOptions | None = None) -> None:
        super().__init__(options)

        # Reverse the list of tags to ensure that the highest level tags are matched first
        emojis_in_precedence_order = list(self.options.tag_to_level.keys())[::-1]

        try:
            highest_emoji_pattern = regexp(
                r"(?P<type>%s)" % str.join("|", emojis_in_precedence_order)
            )
        except re.error as err:
            raise InvalidParserOptions(
                str.join(
                    "\n",
                    [
                        f"Invalid options for {self.__class__.__name__}",
                        "Unable to create regular expression from configured commit-types.",
                        "Please check the configured commit-types and remove or escape any regular expression characters.",
                    ],
                )
            ) from err

        self.emoji_selector = regexp(
            str.join(
                "",
                [
                    f"^{highest_emoji_pattern.pattern}",
                    r"(?:\((?P<scope>[^)]+)\))?:?",
                ],
            )
        )

        # GitHub & Gitea use (#123), GitLab uses (!123), and BitBucket uses (pull request #123)
        self.mr_selector = regexp(
            r"[\t ]+\((?:pull request )?(?P<mr_number>[#!]\d+)\)[\t ]*$"
        )

        self.issue_selector = regexp(
            str.join(
                "",
                [
                    r"^(?:clos(?:e|es|ed|ing)|fix(?:es|ed|ing)?|resolv(?:e|es|ed|ing)|implement(?:s|ed|ing)?):",
                    r"[\t ]+(?P<issue_predicate>.+)[\t ]*$",
                ],
            ),
            flags=re.MULTILINE | re.IGNORECASE,
        )

        self.notice_selector = regexp(r"^NOTICE: (?P<notice>.+)$")

        self.filters = {
            "typo-extra-spaces": (regexp(r"(\S)  +(\S)"), r"\1 \2"),
            "git-header-commit": (
                regexp(r"^[\t ]*commit [0-9a-f]+$\n?", flags=re.MULTILINE),
                "",
            ),
            "git-header-author": (
                regexp(r"^[\t ]*Author: .+$\n?", flags=re.MULTILINE),
                "",
            ),
            "git-header-date": (
                regexp(r"^[\t ]*Date: .+$\n?", flags=re.MULTILINE),
                "",
            ),
            "git-squash-heading": (
                regexp(
                    r"^[\t ]*Squashed commit of the following:.*$\n?",
                    flags=re.MULTILINE,
                ),
                "",
            ),
            "git-squash-commit-prefix": (
                regexp(
                    str.join(
                        "",
                        [
                            r"^(?:[\t ]*[*-][\t ]+|[\t ]+)?",  # bullet points or indentation
                            highest_emoji_pattern.pattern
                            + r"(\W)",  # prior to commit type
                        ],
                    ),
                    flags=re.MULTILINE,
                ),
                # move commit type to the start of the line
                r"\1\2",
            ),
        }

    @staticmethod
    def get_default_options() -> EmojiParserOptions:
        return EmojiParserOptions()

    def commit_body_components_separator(
        self, accumulator: dict[str, list[str]], text: str
    ) -> dict[str, list[str]]:
        if (match := self.notice_selector.match(text)) and (
            notice := match.group("notice")
        ):
            accumulator["notices"].append(notice)
            return accumulator

        if self.options.parse_linked_issues and (
            match := self.issue_selector.search(text)
        ):
            predicate = regexp(r",? and | *[,;/& ] *").sub(
                ",", match.group("issue_predicate") or ""
            )
            # Almost all issue trackers use a number to reference an issue so
            # we use a simple regexp to validate the existence of a number which helps filter out
            # any non-issue references that don't fit our expected format
            has_number = regexp(r"\d+")
            new_issue_refs: set[str] = set(
                filter(
                    lambda issue_str, validator=has_number: validator.search(issue_str),  # type: ignore[arg-type]
                    predicate.split(","),
                )
            )
            if new_issue_refs:
                accumulator["linked_issues"] = sort_numerically(
                    set(accumulator["linked_issues"]).union(new_issue_refs)
                )
                return accumulator

        # Prevent appending duplicate descriptions
        if text not in accumulator["descriptions"]:
            accumulator["descriptions"].append(text)

        return accumulator

    def parse_message(self, message: str) -> ParsedMessageResult:
        msg_parts = message.split("\n", maxsplit=1)
        subject = msg_parts[0]
        msg_body = msg_parts[1] if len(msg_parts) > 1 else ""

        linked_merge_request = ""
        if mr_match := self.mr_selector.search(subject):
            linked_merge_request = mr_match.group("mr_number")
            subject = self.mr_selector.sub("", subject).strip()

        # Search for emoji of the highest importance in the subject
        match = self.emoji_selector.search(subject)
        primary_emoji = match.group("type") if match else "Other"
        parsed_scope = (match.group("scope") if match else None) or ""

        level_bump = self.options.tag_to_level.get(
            primary_emoji, self.options.default_bump_level
        )

        # All emojis will remain part of the returned description
        body_components: dict[str, list[str]] = reduce(
            self.commit_body_components_separator,
            [
                subject,
                *parse_paragraphs(msg_body),
            ],
            {
                "descriptions": [],
                "notices": [],
                "linked_issues": [],
            },
        )

        descriptions = tuple(body_components["descriptions"])

        return ParsedMessageResult(
            bump=level_bump,
            type=primary_emoji,
            category=primary_emoji,
            scope=parsed_scope,
            descriptions=(
                descriptions[:1] if level_bump is LevelBump.MAJOR else descriptions
            ),
            breaking_descriptions=(
                descriptions[1:] if level_bump is LevelBump.MAJOR else ()
            ),
            release_notices=tuple(body_components["notices"]),
            linked_issues=tuple(body_components["linked_issues"]),
            linked_merge_request=linked_merge_request,
        )

    @staticmethod
    def is_merge_commit(commit: Commit) -> bool:
        return len(commit.parents) > 1

    def parse_commit(self, commit: Commit) -> ParseResult:
        return ParsedCommit.from_parsed_message_result(
            commit, self.parse_message(force_str(commit.message))
        )

    def parse(self, commit: Commit) -> ParseResult | list[ParseResult]:
        """
        Parse a commit message

        If the commit message is a squashed merge commit, it will be split into
        multiple commits, each of which will be parsed separately. Single commits
        will be returned as a list of a single ParseResult.
        """
        if self.options.ignore_merge_commits and self.is_merge_commit(commit):
            err_msg = "Ignoring merge commit: %s" % commit.hexsha[:8]
            logger.debug(err_msg)
            return ParseError(commit, err_msg)

        separate_commits: list[Commit] = (
            self.unsquash_commit(commit)
            if self.options.parse_squash_commits
            else [commit]
        )

        # Parse each commit individually if there were more than one
        parsed_commits: list[ParseResult] = list(
            map(self.parse_commit, separate_commits)
        )

        def add_linked_merge_request(
            parsed_result: ParseResult, mr_number: str
        ) -> ParseResult:
            return (
                parsed_result
                if not isinstance(parsed_result, ParsedCommit)
                else ParsedCommit(
                    **{
                        **parsed_result._asdict(),
                        "linked_merge_request": mr_number,
                    }
                )
            )

        # TODO: improve this for other VCS systems other than GitHub & BitBucket
        # Github works as the first commit in a squash merge commit has the PR number
        # appended to the first line of the commit message
        lead_commit = next(iter(parsed_commits))

        if isinstance(lead_commit, ParsedCommit) and lead_commit.linked_merge_request:
            # If the first commit has linked merge requests, assume all commits
            # are part of the same PR and add the linked merge requests to all
            # parsed commits
            parsed_commits = [
                lead_commit,
                *map(
                    lambda parsed_result, mr=lead_commit.linked_merge_request: (  # type: ignore[misc]
                        add_linked_merge_request(parsed_result, mr)
                    ),
                    parsed_commits[1:],
                ),
            ]

        return parsed_commits

    def unsquash_commit(self, commit: Commit) -> list[Commit]:
        # GitHub EXAMPLE:
        # ✨(changelog): add autofit_text_width filter to template environment (#1062)
        #
        # This change adds an equivalent style formatter that can apply a text alignment
        # to a maximum width and also maintain an indent over paragraphs of text
        #
        # * 🌐 Support Japanese language
        #
        # * ✅(changelog-context): add test cases to check autofit_text_width filter use
        #
        # `git merge --squash` EXAMPLE:
        # Squashed commit of the following:
        #
        # commit 63ec09b9e844e616dcaa7bae35a0b66671b59fbb
        # Author: codejedi365 <codejedi365@gmail.com>
        # Date:   Sun Oct 13 12:05:23 2024 -0000
        #
        #     ⚡️ (homepage): Lazyload home screen images
        #
        #
        # Return a list of artificial commits (each with a single commit message)
        return [
            # create a artificial commit object (copy of original but with modified message)
            Commit(
                **{
                    **deep_copy_commit(commit),
                    "message": commit_msg,
                }
            )
            for commit_msg in self.unsquash_commit_message(force_str(commit.message))
        ] or [commit]

    def unsquash_commit_message(self, message: str) -> list[str]:
        normalized_message = message.replace("\r", "").strip()

        # split by obvious separate commits (applies to manual git squash merges)
        obvious_squashed_commits = self.filters["git-header-commit"][0].split(
            normalized_message
        )

        separate_commit_msgs: list[str] = reduce(
            lambda all_msgs, msgs: all_msgs + msgs,
            map(self._find_squashed_commits_in_str, obvious_squashed_commits),
            [],
        )

        return list(filter(None, separate_commit_msgs))

    def _find_squashed_commits_in_str(self, message: str) -> list[str]:
        separate_commit_msgs: list[str] = []
        current_msg = ""

        for paragraph in filter(None, message.strip().split("\n\n")):
            # Apply filters to normalize the paragraph
            clean_paragraph = reduce(text_reducer, self.filters.values(), paragraph)

            # remove any filtered (and now empty) paragraphs (ie. the git headers)
            if not clean_paragraph.strip():
                continue

            # Check if the paragraph is the start of a new emoji commit
            if not self.emoji_selector.search(clean_paragraph):
                if not separate_commit_msgs and not current_msg:
                    # if there are no separate commit messages and no current message
                    # then this is the first commit message
                    current_msg = dedent(clean_paragraph)
                    continue

                # append the paragraph as part of the previous commit message
                if current_msg:
                    current_msg += f"\n\n{dedent(clean_paragraph)}"
                # else: drop the paragraph
                continue

            # Since we found the start of the new commit, store any previous commit
            # message separately and start the new commit message
            if current_msg:
                separate_commit_msgs.append(current_msg)

            current_msg = clean_paragraph

        return [*separate_commit_msgs, current_msg]



from __future__ import annotations

from datetime import timedelta
from itertools import count
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, cast

import pytest

from semantic_release.cli.config import ChangelogOutputFormat
from semantic_release.commit_parser.conventional.options_monorepo import (
    ConventionalCommitMonorepoParserOptions,
)
from semantic_release.commit_parser.conventional.parser_monorepo import (
    ConventionalCommitMonorepoParser,
)
from semantic_release.version.version import Version

import tests.conftest
import tests.const
import tests.util
from tests.const import (
    DEFAULT_BRANCH_NAME,
    EXAMPLE_HVCS_DOMAIN,
    INITIAL_COMMIT_MESSAGE,
    RepoActionStep,
)

if TYPE_CHECKING:
    from typing import Sequence

    from semantic_release.commit_parser._base import CommitParser, ParserOptions
    from semantic_release.commit_parser.token import ParseResult

    from tests.conftest import (
        GetCachedRepoDataFn,
        GetMd5ForSetOfFilesFn,
        GetStableDateNowFn,
    )
    from tests.fixtures.example_project import ExProjectDir
    from tests.fixtures.git_repo import (
        BuildRepoFromDefinitionFn,
        BuildRepoOrCopyCacheFn,
        BuildSpecificRepoFn,
        BuiltRepoResult,
        CommitConvention,
        ConvertCommitSpecsToCommitDefsFn,
        ConvertCommitSpecToCommitDefFn,
        ExProjectGitRepoFn,
        FormatGitHubMergeCommitMsgFn,
        GetRepoDefinitionFn,
        RepoActionChangeDirectory,
        RepoActionGitMerge,
        RepoActionGitMergeDetails,
        RepoActions,
        RepoActionWriteChangelogsDestFile,
        TomlSerializableTypes,
    )


@pytest.fixture(scope="session")
def deps_files_4_github_flow_monorepo_w_feature_release_channel(
    deps_files_4_example_git_monorepo: list[Path],
) -> list[Path]:
    return [
        *deps_files_4_example_git_monorepo,
        # This file
        Path(__file__).absolute(),
        # because of imports
        Path(tests.const.__file__).absolute(),
        Path(tests.util.__file__).absolute(),
        # because of the fixtures
        Path(tests.conftest.__file__).absolute(),
    ]


@pytest.fixture(scope="session")
def build_spec_hash_4_github_flow_monorepo_w_feature_release_channel(
    get_md5_for_set_of_files: GetMd5ForSetOfFilesFn,
    deps_files_4_github_flow_monorepo_w_feature_release_channel: list[Path],
) -> str:
    # Generates a hash of the build spec to set when to invalidate the cache
    return get_md5_for_set_of_files(
        deps_files_4_github_flow_monorepo_w_feature_release_channel
    )


@pytest.fixture(scope="session")
def get_repo_definition_4_github_flow_monorepo_w_feature_release_channel(
    convert_commit_specs_to_commit_defs: ConvertCommitSpecsToCommitDefsFn,
    convert_commit_spec_to_commit_def: ConvertCommitSpecToCommitDefFn,
    format_merge_commit_msg_github: FormatGitHubMergeCommitMsgFn,
    monorepo_pkg1_changelog_md_file: Path,
    monorepo_pkg1_changelog_rst_file: Path,
    monorepo_pkg2_changelog_md_file: Path,
    monorepo_pkg2_changelog_rst_file: Path,
    monorepo_pkg1_name: str,
    monorepo_pkg2_name: str,
    monorepo_pkg1_dir: str,
    monorepo_pkg2_dir: str,
    monorepo_pkg1_docs_dir: str,
    monorepo_pkg2_docs_dir: str,
    monorepo_pkg1_version_py_file: Path,
    monorepo_pkg2_version_py_file: Path,
    stable_now_date: GetStableDateNowFn,
    default_tag_format_str: str,
) -> GetRepoDefinitionFn:
    """
    Builds a Monorepo with the GitHub Flow branching strategy and a merge commit merging strategy
    for alpha feature releases and official releases on the default branch.

    Implementation:
    - The monorepo contains two packages, each with its own internal changelog but shared template.
    - The repository implements the following git graph:

    ```
    * chore(release): pkg2@1.1.0 [skip ci] (tag: pkg2-v1.1.0)
    * Merge pull request #3 from 'pkg2/feat/pr-2'
    |\
    | * chore(release): pkg2@1.1.0-alpha.2 [skip ci] (tag: pkg2-v1.1.0-alpha.2, branch: pkg2/feat/pr-2)
    | * fix(pkg2-cli): file modified outside of pkg 2, identified by scope
    | * chore(release): pkg2@1.1.0-alpha.1 [skip ci] (tag: pkg2-v1.1.0-alpha.1)
    | * docs: pkg2 docs modified outside of pkg 2, identified by path filter
    | * test: add cli tests
    | * feat: no pkg scope but file in pkg 2 directory
    |/
    * chore(release): pkg1@1.0.1 [skip ci] (tag: pkg1-v1.0.1)
    * Merge pull request #2 from 'pkg1/fix/pr-1'
    |\
    | * chore(release): pkg1@1.0.1-alpha.2 [skip ci] (tag: pkg1-v1.0.1-alpha.2, branch: pkg1/fix/pr-1)
    | * fix(pkg1-cli): file modified outside of pkg 1, identified by scope
    | * chore(release): pkg1@1.0.1-alpha.1 [skip ci] (tag: pkg1-v1.0.1-alpha.1)
    | * fix: no pkg scope but file in pkg 1 directory
    |/
    * chore(release): pkg2@1.0.0 [skip ci] (tag: pkg2-v1.0.0)          # Initial release of pkg 2
    * chore(release): pkg1@1.0.0 [skip ci] (tag: pkg1-v1.0.0)          # Initial release of pkg 1
    * Initial commit                                                    # Includes core functionality for both packages
    ```
    """

    def _get_repo_from_definition(
        commit_type: CommitConvention,
        hvcs_client_name: str = "github",
        hvcs_domain: str = EXAMPLE_HVCS_DOMAIN,
        tag_format_str: str | None = default_tag_format_str,
        extra_configs: dict[str, TomlSerializableTypes] | None = None,
        mask_initial_release: bool = True,
        ignore_merge_commits: bool = True,
    ) -> Sequence[RepoActions]:
        stable_now_datetime = stable_now_date()
        commit_timestamp_gen = (
            (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
            for i in count(step=1)
        )
        pr_num_gen = (i for i in count(start=2, step=1))

        pkg1_changelog_file_definitions: Sequence[RepoActionWriteChangelogsDestFile] = [
            {
                "path": monorepo_pkg1_changelog_md_file,
                "format": ChangelogOutputFormat.MARKDOWN,
                "mask_initial_release": True,
            },
            {
                "path": monorepo_pkg1_changelog_rst_file,
                "format": ChangelogOutputFormat.RESTRUCTURED_TEXT,
                "mask_initial_release": True,
            },
        ]

        pkg2_changelog_file_definitions: Sequence[RepoActionWriteChangelogsDestFile] = [
            {
                "path": monorepo_pkg2_changelog_md_file,
                "format": ChangelogOutputFormat.MARKDOWN,
                "mask_initial_release": True,
            },
            {
                "path": monorepo_pkg2_changelog_rst_file,
                "format": ChangelogOutputFormat.RESTRUCTURED_TEXT,
                "mask_initial_release": True,
            },
        ]

        change_to_pkg1_dir: RepoActionChangeDirectory = {
            "action": RepoActionStep.CHANGE_DIRECTORY,
            "details": {
                "directory": monorepo_pkg1_dir,
            },
        }

        change_to_pkg2_dir: RepoActionChangeDirectory = {
            "action": RepoActionStep.CHANGE_DIRECTORY,
            "details": {
                "directory": monorepo_pkg2_dir,
            },
        }

        change_to_example_project_dir: RepoActionChangeDirectory = {
            "action": RepoActionStep.CHANGE_DIRECTORY,
            "details": {
                "directory": "/",
            },
        }

        if commit_type != "conventional":
            raise ValueError(f"Unsupported commit type: {commit_type}")

        pkg1_path_filters = (".", f"../../{monorepo_pkg1_docs_dir}")
        pkg1_commit_parser = ConventionalCommitMonorepoParser(
            options=ConventionalCommitMonorepoParserOptions(
                parse_squash_commits=True,
                ignore_merge_commits=ignore_merge_commits,
                scope_prefix=f"{monorepo_pkg1_name}-?",
                path_filters=pkg1_path_filters,
            )
        )

        pkg2_path_filters = (".", f"../../{monorepo_pkg2_docs_dir}")
        pkg2_commit_parser = ConventionalCommitMonorepoParser(
            options=ConventionalCommitMonorepoParserOptions(
                parse_squash_commits=pkg1_commit_parser.options.parse_squash_commits,
                ignore_merge_commits=pkg1_commit_parser.options.ignore_merge_commits,
                scope_prefix=f"{monorepo_pkg2_name}-?",
                path_filters=pkg2_path_filters,
            )
        )

        common_configs: dict[str, TomlSerializableTypes] = {
            # Set the default release branch
            "tool.semantic_release.branches.main": {
                "match": r"^(main|master)$",
                "prerelease": False,
            },
            "tool.semantic_release.allow_zero_version": False,
            "tool.semantic_release.changelog.exclude_commit_patterns": [r"^chore"],
            "tool.semantic_release.commit_parser": f"{commit_type}-monorepo",
            "tool.semantic_release.commit_parser_options.parse_squash_commits": pkg1_commit_parser.options.parse_squash_commits,
            "tool.semantic_release.commit_parser_options.ignore_merge_commits": pkg1_commit_parser.options.ignore_merge_commits,
        }

        mr1_pkg1_fix_branch_name = f"{monorepo_pkg1_name}/fix/pr-1"
        mr2_pkg2_feat_branch_name = f"{monorepo_pkg2_name}/feat/pr-2"

        pkg1_new_version = Version.parse(
            "1.0.0", tag_format=f"{monorepo_pkg1_name}-{tag_format_str}"
        )
        pkg2_new_version = Version.parse(
            "1.0.0", tag_format=f"{monorepo_pkg2_name}-{tag_format_str}"
        )

        repo_construction_steps: list[RepoActions] = [
            {
                "action": RepoActionStep.CREATE_MONOREPO,
                "details": {
                    "commit_type": commit_type,
                    "hvcs_client_name": hvcs_client_name,
                    "hvcs_domain": hvcs_domain,
                    "post_actions": [
                        {
                            "action": RepoActionStep.CONFIGURE_MONOREPO,
                            "details": {
                                "package_dir": monorepo_pkg1_dir,
                                "package_name": monorepo_pkg1_name,
                                "tag_format_str": pkg1_new_version.tag_format,
                                "mask_initial_release": mask_initial_release,
                                "extra_configs": {
                                    **common_configs,
                                    "tool.semantic_release.commit_message": (
                                        pkg1_cmt_msg_format := dedent(
                                            f"""\
                                            chore(release): {monorepo_pkg1_name}@{{version}} [skip ci]

                                            Automatically generated by python-semantic-release
                                            """
                                        )
                                    ),
                                    # package branches "feat/" & "fix/" has prerelease suffix of "alpha"
                                    "tool.semantic_release.branches.alpha-release": {
                                        "match": rf"^{monorepo_pkg1_name}/(feat|fix)/.+",
                                        "prerelease": True,
                                        "prerelease_token": "alpha",
                                    },
                                    "tool.semantic_release.commit_parser_options.scope_prefix": pkg1_commit_parser.options.scope_prefix,
                                    "tool.semantic_release.commit_parser_options.path_filters": pkg1_path_filters,
                                    **(extra_configs or {}),
                                },
                            },
                        },
                        {
                            "action": RepoActionStep.CONFIGURE_MONOREPO,
                            "details": {
                                "package_dir": monorepo_pkg2_dir,
                                "package_name": monorepo_pkg2_name,
                                "tag_format_str": pkg2_new_version.tag_format,
                                "mask_initial_release": mask_initial_release,
                                "extra_configs": {
                                    **common_configs,
                                    "tool.semantic_release.commit_message": (
                                        pkg2_cmt_msg_format := dedent(
                                            f"""\
                                        chore(release): {monorepo_pkg2_name}@{{version}} [skip ci]

                                        Automatically generated by python-semantic-release
                                        """
                                        )
                                    ),
                                    # package branches "feat/" & "fix/" has prerelease suffix of "alpha"
                                    "tool.semantic_release.branches.alpha-release": {
                                        "match": rf"^{monorepo_pkg2_name}/(feat|fix)/.+",
                                        "prerelease": True,
                                        "prerelease_token": "alpha",
                                    },
                                    "tool.semantic_release.commit_parser_options.scope_prefix": pkg2_commit_parser.options.scope_prefix,
                                    "tool.semantic_release.commit_parser_options.path_filters": pkg2_path_filters,
                                    **(extra_configs or {}),
                                },
                            },
                        },
                        {
                            "action": RepoActionStep.MAKE_COMMITS,
                            "details": {
                                "commits": convert_commit_specs_to_commit_defs(
                                    [
                                        {
                                            "cid": (
                                                cid_c1_initial := "c1_initial_commit"
                                            ),
                                            "conventional": INITIAL_COMMIT_MESSAGE,
                                            "emoji": INITIAL_COMMIT_MESSAGE,
                                            "scipy": INITIAL_COMMIT_MESSAGE,
                                            "datetime": next(commit_timestamp_gen),
                                            "include_in_changelog": bool(
                                                commit_type == "emoji"
                                            ),
                                        },
                                    ],
                                    commit_type,
                                    # this parser does not matter since the commit is common
                                    parser=cast(
                                        "CommitParser[ParseResult, ParserOptions]",
                                        pkg1_commit_parser,
                                    ),
                                    monorepo=True,
                                ),
                            },
                        },
                    ],
                },
            }
        ]

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(pkg1_new_version),
                        "datetime": next(commit_timestamp_gen),
                        "tag_format": pkg1_new_version.tag_format,
                        "version_py_file": monorepo_pkg1_version_py_file.relative_to(
                            monorepo_pkg1_dir
                        ),
                        "commit_message_format": pkg1_cmt_msg_format,
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": pkg1_new_version,
                                    "dest_files": pkg1_changelog_file_definitions,
                                    "commit_ids": [cid_c1_initial],
                                },
                            },
                            change_to_pkg1_dir,
                        ],
                        "post_actions": [change_to_example_project_dir],
                    },
                },
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(pkg2_new_version),
                        "datetime": next(commit_timestamp_gen),
                        "tag_format": pkg2_new_version.tag_format,
                        "version_py_file": monorepo_pkg2_version_py_file.relative_to(
                            monorepo_pkg2_dir
                        ),
                        "commit_message_format": pkg2_cmt_msg_format,
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": pkg2_new_version,
                                    "dest_files": pkg2_changelog_file_definitions,
                                    "commit_ids": [cid_c1_initial],
                                },
                            },
                            change_to_pkg2_dir,
                        ],
                        "post_actions": [change_to_example_project_dir],
                    },
                },
            ]
        )

        # Make a fix in package 1 and release it as an alpha release
        pkg1_new_version = Version.parse(
            "1.0.1-alpha.1", tag_format=pkg1_new_version.tag_format
        )

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {
                        "create_branch": {
                            "name": mr1_pkg1_fix_branch_name,
                            "start_branch": DEFAULT_BRANCH_NAME,
                        }
                    },
                },
                {
                    "action": RepoActionStep.MAKE_COMMITS,
                    "details": {
                        "pre_actions": [change_to_pkg1_dir],
                        "commits": convert_commit_specs_to_commit_defs(
                            [
                                {
                                    "cid": (
                                        cid_pkg1_fib1_c1_fix
                                        := "pkg1_fix_branch_1_c1_fix"
                                    ),
                                    "conventional": "fix: no pkg scope but file in pkg 1 directory\n\nResolves: #123\n",
                                    "emoji": ":bug: no pkg scope but file in pkg 1 directory\n\nResolves: #123\n",
                                    "scipy": "MAINT: no pkg scope but file in pkg 1 directory\n\nResolves: #123\n",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                            ],
                            commit_type,
                            parser=cast(
                                "CommitParser[ParseResult, ParserOptions]",
                                pkg1_commit_parser,
                            ),
                            monorepo=True,
                        ),
                        "post_actions": [change_to_example_project_dir],
                    },
                },
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(pkg1_new_version),
                        "datetime": next(commit_timestamp_gen),
                        "tag_format": pkg1_new_version.tag_format,
                        "version_py_file": monorepo_pkg1_version_py_file.relative_to(
                            monorepo_pkg1_dir
                        ),
                        "commit_message_format": pkg1_cmt_msg_format,
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": pkg1_new_version,
                                    "dest_files": pkg1_changelog_file_definitions,
                                    "commit_ids": [cid_pkg1_fib1_c1_fix],
                                },
                            },
                            change_to_pkg1_dir,
                        ],
                        "post_actions": [change_to_example_project_dir],
                    },
                },
            ]
        )

        # Update the fix in package 1 and release another alpha release
        pkg1_new_version = Version.parse(
            "1.0.1-alpha.2", tag_format=pkg1_new_version.tag_format
        )

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.MAKE_COMMITS,
                    "details": {
                        "commits": convert_commit_specs_to_commit_defs(
                            [
                                {
                                    "cid": (
                                        cid_pkg1_fib1_c2_fix
                                        := "pkg1_fix_branch_1_c2_fix"
                                    ),
                                    "conventional": "fix(pkg1-cli): file modified outside of pkg 1, identified by scope\n\n",
                                    "emoji": ":bug: (pkg1-cli) file modified outside of pkg 1, identified by scope",
                                    "scipy": "MAINT:pkg1-cli: file modified outside of pkg 1, identified by scope",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                            ],
                            commit_type,
                            parser=cast(
                                "CommitParser[ParseResult, ParserOptions]",
                                pkg1_commit_parser,
                            ),
                            monorepo=True,
                        ),
                    },
                },
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(pkg1_new_version),
                        "datetime": next(commit_timestamp_gen),
                        "tag_format": pkg1_new_version.tag_format,
                        "version_py_file": monorepo_pkg1_version_py_file.relative_to(
                            monorepo_pkg1_dir
                        ),
                        "commit_message_format": pkg1_cmt_msg_format,
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": pkg1_new_version,
                                    "dest_files": pkg1_changelog_file_definitions,
                                    "commit_ids": [cid_pkg1_fib1_c2_fix],
                                },
                            },
                            change_to_pkg1_dir,
                        ],
                        "post_actions": [change_to_example_project_dir],
                    },
                },
            ]
        )

        # Merge the fix branch into the default branch and formally release it
        pkg1_new_version = Version.parse(
            "1.0.1", tag_format=pkg1_new_version.tag_format
        )

        merge_def_type_placeholder: RepoActionGitMerge[RepoActionGitMergeDetails] = {
            "action": RepoActionStep.GIT_MERGE,
            "details": {
                "branch_name": mr1_pkg1_fix_branch_name,
                "fast_forward": False,
                "commit_def": convert_commit_spec_to_commit_def(
                    {
                        "cid": (cid_pkg1_fib1_merge := "pkg1_fix_branch_1_merge"),
                        "conventional": (
                            merge_msg := format_merge_commit_msg_github(
                                pr_number=next(pr_num_gen),
                                branch_name=mr1_pkg1_fix_branch_name,
                            )
                        ),
                        "emoji": merge_msg,
                        "scipy": merge_msg,
                        "datetime": next(commit_timestamp_gen),
                        "include_in_changelog": not ignore_merge_commits,
                    },
                    commit_type,
                    parser=cast(
                        "CommitParser[ParseResult, ParserOptions]",
                        pkg1_commit_parser,
                    ),
                    monorepo=True,
                ),
            },
        }

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": DEFAULT_BRANCH_NAME},
                },
                merge_def_type_placeholder,
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(pkg1_new_version),
                        "datetime": next(commit_timestamp_gen),
                        "tag_format": pkg1_new_version.tag_format,
                        "version_py_file": monorepo_pkg1_version_py_file.relative_to(
                            monorepo_pkg1_dir
                        ),
                        "commit_message_format": pkg1_cmt_msg_format,
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": pkg1_new_version,
                                    "dest_files": pkg1_changelog_file_definitions,
                                    "commit_ids": [cid_pkg1_fib1_merge],
                                },
                            },
                            change_to_pkg1_dir,
                        ],
                        "post_actions": [change_to_example_project_dir],
                    },
                },
            ]
        )

        # Make a feature branch and release it as an alpha release
        pkg2_new_version = Version.parse(
            "1.1.0-alpha.1", tag_format=pkg2_new_version.tag_format
        )

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {
                        "create_branch": {
                            "name": mr2_pkg2_feat_branch_name,
                            "start_branch": DEFAULT_BRANCH_NAME,
                        },
                    },
                },
                {
                    "action": RepoActionStep.MAKE_COMMITS,
                    "details": {
                        "pre_actions": [change_to_pkg2_dir],
                        "commits": convert_commit_specs_to_commit_defs(
                            [
                                {
                                    "cid": (
                                        cid_pkg2_feb1_c1_feat
                                        := "pkg2_feat_branch_1_c1_feat"
                                    ),
                                    "conventional": "feat: no pkg scope but file in pkg 2 directory\n",
                                    "emoji": ":sparkles: no pkg scope but file in pkg 2 directory",
                                    "scipy": "ENH: no pkg scope but file in pkg 2 directory",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                                {
                                    "cid": (
                                        cid_pkg2_feb1_c2_test
                                        := "pkg2_feat_branch_1_c2_test"
                                    ),
                                    "conventional": "test: add cli tests",
                                    "emoji": ":checkmark: add cli tests",
                                    "scipy": "TST: add cli tests",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                                {
                                    "cid": (
                                        cid_pkg2_feb1_c3_docs
                                        := "pkg2_feat_branch_1_c3_docs"
                                    ),
                                    "conventional": "docs: pkg2 docs modified outside of pkg 2, identified by path filter",
                                    "emoji": ":book: pkg2 docs modified outside of pkg 2, identified by path filter",
                                    "scipy": "DOC: pkg2 docs modified outside of pkg 2, identified by path filter",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                    "file_to_change": f"../../{monorepo_pkg2_docs_dir}/index.rst",
                                },
                            ],
                            commit_type,
                            parser=cast(
                                "CommitParser[ParseResult, ParserOptions]",
                                pkg2_commit_parser,
                            ),
                            monorepo=True,
                        ),
                        "post_actions": [change_to_example_project_dir],
                    },
                },
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(pkg2_new_version),
                        "datetime": next(commit_timestamp_gen),
                        "tag_format": pkg2_new_version.tag_format,
                        "version_py_file": monorepo_pkg2_version_py_file.relative_to(
                            monorepo_pkg2_dir
                        ),
                        "commit_message_format": pkg2_cmt_msg_format,
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": pkg2_new_version,
                                    "dest_files": pkg2_changelog_file_definitions,
                                    "commit_ids": [
                                        cid_pkg2_feb1_c1_feat,
                                        cid_pkg2_feb1_c2_test,
                                        cid_pkg2_feb1_c3_docs,
                                    ],
                                },
                            },
                            change_to_pkg2_dir,
                        ],
                        "post_actions": [change_to_example_project_dir],
                    },
                },
            ]
        )

        # Update the feat with a fix in package 2 and release another alpha release
        pkg2_new_version = Version.parse(
            "1.1.0-alpha.2", tag_format=pkg2_new_version.tag_format
        )

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.MAKE_COMMITS,
                    "details": {
                        "commits": convert_commit_specs_to_commit_defs(
                            [
                                {
                                    "cid": (
                                        cid_pkg2_feb1_c4_fix
                                        := "pkg2_feat_branch_1_c4_fix"
                                    ),
                                    "conventional": "fix(pkg2-cli): file modified outside of pkg 2, identified by scope",
                                    "emoji": ":bug: (pkg2-cli) file modified outside of pkg 2, identified by scope",
                                    "scipy": "MAINT:pkg2-cli: file modified outside of pkg 2, identified by scope",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                            ],
                            commit_type,
                            parser=cast(
                                "CommitParser[ParseResult, ParserOptions]",
                                pkg2_commit_parser,
                            ),
                            monorepo=True,
                        ),
                    },
                },
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(pkg2_new_version),
                        "datetime": next(commit_timestamp_gen),
                        "tag_format": pkg2_new_version.tag_format,
                        "version_py_file": monorepo_pkg2_version_py_file.relative_to(
                            monorepo_pkg2_dir
                        ),
                        "commit_message_format": pkg2_cmt_msg_format,
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": pkg2_new_version,
                                    "dest_files": pkg2_changelog_file_definitions,
                                    "commit_ids": [cid_pkg2_feb1_c4_fix],
                                },
                            },
                            change_to_pkg2_dir,
                        ],
                        "post_actions": [change_to_example_project_dir],
                    },
                },
            ]
        )

        # Merge the feat branch into the default branch and formally release a package 2
        pkg2_new_version = Version.parse(
            "1.1.0", tag_format=pkg2_new_version.tag_format
        )

        merge_def_type_placeholder = {
            "action": RepoActionStep.GIT_MERGE,
            "details": {
                "branch_name": mr2_pkg2_feat_branch_name,
                "fast_forward": False,
                "commit_def": convert_commit_spec_to_commit_def(
                    {
                        "cid": (cid_pkg2_feb1_merge := "pkg2_feat_branch_1_merge"),
                        "conventional": (
                            merge_msg := format_merge_commit_msg_github(
                                pr_number=next(pr_num_gen),
                                branch_name=mr2_pkg2_feat_branch_name,
                            )
                        ),
                        "emoji": merge_msg,
                        "scipy": merge_msg,
                        "datetime": next(commit_timestamp_gen),
                        "include_in_changelog": not ignore_merge_commits,
                    },
                    commit_type,
                    parser=cast(
                        "CommitParser[ParseResult, ParserOptions]",
                        pkg2_commit_parser,
                    ),
                    monorepo=True,
                ),
            },
        }

        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.GIT_CHECKOUT,
                    "details": {"branch": DEFAULT_BRANCH_NAME},
                },
                merge_def_type_placeholder,
                {
                    "action": RepoActionStep.RELEASE,
                    "details": {
                        "version": str(pkg2_new_version),
                        "datetime": next(commit_timestamp_gen),
                        "tag_format": pkg2_new_version.tag_format,
                        "version_py_file": monorepo_pkg2_version_py_file.relative_to(
                            monorepo_pkg2_dir
                        ),
                        "commit_message_format": pkg2_cmt_msg_format,
                        "pre_actions": [
                            {
                                "action": RepoActionStep.WRITE_CHANGELOGS,
                                "details": {
                                    "new_version": pkg2_new_version,
                                    "dest_files": pkg2_changelog_file_definitions,
                                    "commit_ids": [cid_pkg2_feb1_merge],
                                },
                            },
                            change_to_pkg2_dir,
                        ],
                        "post_actions": [change_to_example_project_dir],
                    },
                },
            ]
        )

        return repo_construction_steps

    return _get_repo_from_definition


@pytest.fixture(scope="session")
def build_monorepo_w_github_flow_w_feature_release_channel(
    build_repo_from_definition: BuildRepoFromDefinitionFn,
    get_repo_definition_4_github_flow_monorepo_w_feature_release_channel: GetRepoDefinitionFn,
    get_cached_repo_data: GetCachedRepoDataFn,
    build_repo_or_copy_cache: BuildRepoOrCopyCacheFn,
    build_spec_hash_4_github_flow_monorepo_w_feature_release_channel: str,
) -> BuildSpecificRepoFn:
    def _build_specific_repo_type(
        repo_name: str, commit_type: CommitConvention, dest_dir: Path
    ) -> Sequence[RepoActions]:
        def _build_repo(cached_repo_path: Path) -> Sequence[RepoActions]:
            repo_construction_steps = (
                get_repo_definition_4_github_flow_monorepo_w_feature_release_channel(
                    commit_type=commit_type,
                )
            )
            return build_repo_from_definition(cached_repo_path, repo_construction_steps)

        build_repo_or_copy_cache(
            repo_name=repo_name,
            build_spec_hash=build_spec_hash_4_github_flow_monorepo_w_feature_release_channel,
            build_repo_func=_build_repo,
            dest_dir=dest_dir,
        )

        if not (cached_repo_data := get_cached_repo_data(proj_dirname=repo_name)):
            raise ValueError("Failed to retrieve repo data from cache")

        return cached_repo_data["build_definition"]

    return _build_specific_repo_type


# --------------------------------------------------------------------------- #
# Test-level fixtures that will cache the built directory & set up test case  #
# --------------------------------------------------------------------------- #


@pytest.fixture
def monorepo_w_github_flow_w_feature_release_channel_conventional_commits(
    build_monorepo_w_github_flow_w_feature_release_channel: BuildSpecificRepoFn,
    example_project_git_repo: ExProjectGitRepoFn,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
) -> BuiltRepoResult:
    repo_name = (
        monorepo_w_github_flow_w_feature_release_channel_conventional_commits.__name__
    )
    commit_type: CommitConvention = repo_name.split("_")[-2]  # type: ignore[assignment]

    return {
        "definition": build_monorepo_w_github_flow_w_feature_release_channel(
            repo_name=repo_name,
            commit_type=commit_type,
            dest_dir=example_project_dir,
        ),
        "repo": example_project_git_repo(),
    }



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest
import tomlkit
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from tests.const import EXAMPLE_PROJECT_NAME, MAIN_PROG_NAME, VERSION_SUBCMD
from tests.fixtures import (
    repo_w_no_tags_conventional_commits,
)
from tests.util import (
    assert_successful_exit_code,
    dynamic_python_import,
)

if TYPE_CHECKING:
    from typing import List
    from unittest.mock import MagicMock

    from requests_mock import Mocker
    from typing_extensions import TypeAlias

    from tests.conftest import RunCliFn
    from tests.fixtures.example_project import (
        ExProjectDir,
        GetExpectedVersionPyFileContentFn,
        UpdatePyprojectTomlFn,
    )
    from tests.fixtures.git_repo import BuiltRepoResult

    CaseId: TypeAlias = str
    CliArgs: TypeAlias = List[str]
    NextReleaseVersion: TypeAlias = str
    ExistingTags: TypeAlias = List[str]
    ExpectedNewPartialTags: TypeAlias = List[str]
    ExpectedMovedPartialTags: TypeAlias = List[str]


cases: tuple[
    tuple[
        CaseId,
        CliArgs,
        NextReleaseVersion,
        ExistingTags,
        ExpectedNewPartialTags,
        ExpectedMovedPartialTags,
    ],
    ...,
] = (
    # pre-release should not affect partial tags
    (
        "pre-release",
        ["--prerelease"],
        "0.0.0-rc.1",
        ["v0", "v0.0"],
        [],
        [],
    ),
    # Create partial tags when they don't exist
    (
        "create-partial-tags-when-they-dont-exist__build-metadata",
        ["--minor", "--build-metadata", "build.12345"],
        "0.1.0+build.12345",
        [],
        ["v0", "v0.1", "v0.1.0"],
        [],
    ),
    (
        "create-partial-tags-when-they-dont-exist__patch",
        ["--patch"],
        "0.0.1",
        [],
        ["v0", "v0.0"],
        [],
    ),
    (
        "create-partial-tags-when-they-dont-exist__minor",
        ["--minor"],
        "0.1.0",
        [],
        ["v0", "v0.1"],
        [],
    ),
    (
        "create-partial-tags-when-they-dont-exist__major",
        ["--major"],
        "1.0.0",
        [],
        ["v1", "v1.0"],
        [],
    ),
    # Update existing partial tags
    (
        "update-existing-partial-tags__build-metadata",
        ["--patch", "--build-metadata", "build.12345"],
        "0.1.1+build.12345",
        ["v0", "v0.0", "v0.1", "v0.1.0"],
        ["v0.1.1"],
        ["v0", "v0.1"],
    ),
    (
        "update-existing-partial-tags__patch",
        ["--patch"],
        "0.0.1",
        ["v0", "v0.0"],
        [],
        ["v0", "v0.0"],
    ),
    (
        "update-existing-partial-tags__minor",
        ["--minor"],
        "0.1.0",
        ["v0", "v0.0", "v0.1"],
        [],
        ["v0", "v0.1"],
    ),
    (
        "update-existing-partial-tags__major",
        ["--major"],
        "1.0.0",
        ["v0", "v0.0", "v0.1", "v1", "v1.0"],
        [],
        ["v1", "v1.0"],
    ),
    # Update existing partial tags and create new one
    (
        "update-existing-partial-tags-and-create-new-one",
        ["--minor"],
        "0.1.0",
        ["v0", "v0.0"],
        ["v0.1"],
        ["v0"],
    ),
    # Partial tag disabled for older version, now enabled
    (
        "partial-tag-disabled-for-older-version__build-metadata",
        ["--patch", "--build-metadata", "build.12345"],
        "1.1.2+build.12345",
        ["v0.1.0", "v0.1.1", "v1.0.0", "v1.1.0", "v1.1.1+build.1234"],
        ["v1", "v1.1", "v1.1.2"],
        [],
    ),
    (
        "partial-tag-disabled-for-older-version__patch",
        ["--patch"],
        "1.1.2",
        ["v0.1.0", "v0.1.1", "v1.0.0", "v1.1.0", "v1.1.1"],
        ["v1", "v1.1"],
        [],
    ),
    (
        "partial-tag-disabled-for-older-version__minor",
        ["--minor"],
        "1.2.0",
        ["v0.1.0", "v0.1.1", "v1.0.0", "v1.1.0", "v1.1.1"],
        ["v1", "v1.2"],
        [],
    ),
    (
        "partial-tag-enabled-for-newer-version__major",
        ["--major"],
        "2.0.0",
        ["v0.1.0", "v0.1.1", "v1.0.0", "v1.1.0", "v1.1.1"],
        ["v2", "v2.0"],
        [],
    ),
)


@pytest.mark.parametrize(
    "repo_result, add_partial_tags, cli_args, next_release_version, existing_partial_tags, expected_new_partial_tags, expected_moved_partial_tags",
    [
        *(
            pytest.param(
                lazy_fixture(repo_w_no_tags_conventional_commits.__name__),
                True,
                cli_args,
                next_release_version,
                existing_tags,
                expected_new_partial_tags,
                expected_moved_partial_tags,
                id=f"{case_id}__partial-tags-enabled",
            )
            for case_id, cli_args, next_release_version, existing_tags, expected_new_partial_tags, expected_moved_partial_tags in cases
        ),
        *(
            pytest.param(
                lazy_fixture(repo_w_no_tags_conventional_commits.__name__),
                False,
                cli_args,
                next_release_version,
                existing_tags,
                expected_new_partial_tags,
                expected_moved_partial_tags,
                id=f"{case_id}__partial-tags-disabled",
            )
            for case_id, cli_args, next_release_version, existing_tags, expected_new_partial_tags, expected_moved_partial_tags in cases
        ),
    ],
)
def test_version_partial_tag_creation(
    repo_result: BuiltRepoResult,
    add_partial_tags: bool,
    cli_args: list[str],
    next_release_version: str,
    example_project_dir: ExProjectDir,
    example_pyproject_toml: Path,
    existing_partial_tags: list[str],
    expected_new_partial_tags: list[str],
    expected_moved_partial_tags: list[str],
    run_cli: RunCliFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_file: Path,
    changelog_md_file: Path,
    version_py_file: Path,
    get_expected_version_py_file_content: GetExpectedVersionPyFileContentFn,
):
    # Force clean directory state before test (needed for the repo_w_no_tags)
    repo = repo_result["repo"]
    repo.git.reset("HEAD", hard=True)

    # Enable partial tags
    update_pyproject_toml("tool.semantic_release.add_partial_tags", add_partial_tags)

    expected_changed_files = sorted(
        [
            str(changelog_md_file),
            str(pyproject_toml_file),
            str(version_py_file),
        ]
    )
    expected_new_partial_tags = expected_new_partial_tags if add_partial_tags else []
    expected_moved_partial_tags = (
        expected_moved_partial_tags if add_partial_tags else []
    )

    expected_version_py_content = get_expected_version_py_file_content(
        next_release_version
    )

    # Setup: create existing tags
    for tag in existing_partial_tags:
        repo.create_tag(tag)

    # Setup: take measurement before running the version command
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name: repo.commit(tag) for tag in repo.tags}

    pyproject_toml_before = tomlkit.loads(
        example_pyproject_toml.read_text(encoding="utf-8")
    )

    # Modify the pyproject.toml to remove the version so we can compare it later
    pyproject_toml_before.get("tool", {}).get("poetry", {}).pop("version", None)

    # Define expectations before execution (hypothesis)
    expected_git_fetch_calls = 1
    expected_vcs_release_calls = 1
    # 1 for commit, 1 for tag, 1 for each moved or created partial tag
    expected_git_push_calls = (
        2 + len(expected_new_partial_tags) + len(expected_moved_partial_tags)
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *cli_args]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    head_after = repo.head.commit
    tags_after = {tag.name: repo.commit(tag.name) for tag in repo.tags}
    new_tags = {tag: sha for tag, sha in tags_after.items() if tag not in tags_before}
    moved_tags = {
        tag: sha
        for tag, sha in tags_after.items()
        if tag in tags_before and sha != tags_before[tag]
    }
    differing_files = sorted(
        [
            # Make sure filepath uses os specific path separators
            str(Path(file))
            for file in str(
                repo.git.diff("HEAD", "HEAD~1", name_only=True)
            ).splitlines()
        ]
    )
    pyproject_toml_after = tomlkit.loads(
        example_pyproject_toml.read_text(encoding="utf-8")
    )
    pyproj_version_after = (
        pyproject_toml_after.get("tool", {}).get("poetry", {}).pop("version")
    )

    # Load python module for reading the version (ensures the file is valid)
    actual_version_py_content = (example_project_dir / version_py_file).read_text()

    # Evaluate (normal release actions should have occurred when forced patch bump)
    assert_successful_exit_code(result, cli_cmd)

    # A commit has been made
    assert [head_sha_before] == [head.hexsha for head in head_after.parents]

    # A version tag and the expected partial tag have been created
    assert 1 + len(expected_new_partial_tags) == len(new_tags)
    assert len(expected_moved_partial_tags) == len(moved_tags)
    assert f"v{next_release_version}" in new_tags

    # Check that all new tags and moved tags are present and on the head commit
    for partial_tag in expected_new_partial_tags:
        assert partial_tag in new_tags
        assert repo.commit(partial_tag).hexsha == head_after.hexsha

    for partial_tag in expected_moved_partial_tags:
        assert partial_tag in moved_tags
        assert repo.commit(partial_tag).hexsha == head_after.hexsha

    # Expected external calls
    assert (
        expected_git_fetch_calls == mocked_git_fetch.call_count
    )  # fetch occurred before push
    assert expected_git_push_calls == mocked_git_push.call_count
    assert (
        expected_vcs_release_calls == post_mocker.call_count
    )  # vcs release creation occurred

    # Changelog already reflects changes this should introduce
    assert expected_changed_files == differing_files

    # Compare pyproject.toml
    assert pyproject_toml_before == pyproject_toml_after
    assert next_release_version == pyproj_version_after

    # Compare _version.py
    assert expected_version_py_content == actual_version_py_content

    # Verify content is parsable & importable
    dynamic_version = dynamic_python_import(
        example_project_dir / version_py_file, f"{EXAMPLE_PROJECT_NAME}._version"
    ).__version__

    assert next_release_version == dynamic_version



from __future__ import annotations

import os
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING

# NOTE: use backport with newer API than stdlib
from importlib_resources import files

import semantic_release
from semantic_release.changelog.context import (
    ReleaseNotesContext,
    autofit_text_width,
    create_pypi_url,
    make_changelog_context,
)
from semantic_release.changelog.template import environment, recursive_render
from semantic_release.cli.config import ChangelogOutputFormat
from semantic_release.cli.const import (
    DEFAULT_CHANGELOG_NAME_STEM,
    DEFAULT_RELEASE_NOTES_TPL_FILE,
    JINJA2_EXTENSION,
)
from semantic_release.cli.util import noop_report
from semantic_release.errors import InternalError
from semantic_release.globals import logger
from semantic_release.helpers import sort_numerically

if TYPE_CHECKING:  # pragma: no cover
    from jinja2 import Environment

    from semantic_release.changelog.context import ChangelogContext
    from semantic_release.changelog.release_history import Release, ReleaseHistory
    from semantic_release.cli.config import RuntimeContext
    from semantic_release.hvcs._base import HvcsBase


def get_default_tpl_dir(style: str, sub_dir: str | None = None) -> Path:
    module_base_path = Path(str(files(semantic_release.__name__)))
    default_templates_path = module_base_path.joinpath(
        f"data/templates/{style}",
        "" if sub_dir is None else sub_dir.strip("/"),
    )

    if default_templates_path.is_dir():
        return default_templates_path

    raise InternalError(
        str.join(
            " ",
            [
                "Default template directory not found at",
                f"{default_templates_path}. Installation corrupted!",
            ],
        )
    )


def render_default_changelog_file(
    output_format: ChangelogOutputFormat,
    changelog_context: ChangelogContext,
    changelog_style: str,
) -> str:
    tpl_dir = get_default_tpl_dir(style=changelog_style, sub_dir=output_format.value)
    changelog_tpl_file = Path(DEFAULT_CHANGELOG_NAME_STEM).with_suffix(
        str.join(".", ["", output_format.value, JINJA2_EXTENSION.lstrip(".")])
    )

    # Create a new environment as we don't want user's configuration as it might
    # not match our default template structure
    template_env = changelog_context.bind_to_environment(
        environment(
            autoescape=False,
            newline_sequence="\n",
            template_dir=tpl_dir,
        )
    )

    # Using the proper enviroment with the changelog context, render the template
    template = template_env.get_template(str(changelog_tpl_file))
    changelog_content = template.render().rstrip()

    # Normalize line endings to ensure universal newlines because that is what is expected
    # of the content when we write it to a file. When using pathlib.Path.write_text(), it
    # will automatically normalize the file to the OS. At this point after render, we may
    # have mixed line endings because of the read_file() call of the previous changelog
    # (which may be /r/n or /n)
    return str.join(
        "\n", [line.replace("\r", "") for line in changelog_content.split("\n")]
    )


def render_release_notes(
    release_notes_template_file: str,
    template_env: Environment,
) -> str:
    # NOTE: release_notes_template_file must be a relative path to the template directory
    # because jinja2's filtering and template loading filter is janky
    template = template_env.get_template(release_notes_template_file)
    release_notes = template.render().rstrip() + os.linesep

    # Normalize line endings to match the current platform
    return str.join(
        os.linesep, [line.replace("\r", "") for line in release_notes.split("\n")]
    )


def apply_user_changelog_template_directory(
    template_dir: Path,
    environment: Environment,
    destination_dir: Path,
    noop: bool = False,
) -> list[str]:
    if noop:
        noop_report(
            str.join(
                " ",
                [
                    "would have recursively rendered the template directory",
                    f"{template_dir!r} relative to {destination_dir!r}.",
                    "Paths which would be modified by this operation cannot be",
                    "determined in no-op mode.",
                ],
            )
        )
        return []

    return recursive_render(
        template_dir, environment=environment, _root_dir=destination_dir
    )


def write_default_changelog(
    changelog_file: Path,
    destination_dir: Path,
    output_format: ChangelogOutputFormat,
    changelog_context: ChangelogContext,
    changelog_style: str,
    noop: bool = False,
) -> str:
    if noop:
        noop_report(
            str.join(
                " ",
                [
                    "would have written your changelog to",
                    str(changelog_file.relative_to(destination_dir)),
                ],
            )
        )
        return str(changelog_file)

    changelog_text = render_default_changelog_file(
        output_format=output_format,
        changelog_context=changelog_context,
        changelog_style=changelog_style,
    )
    # write_text() will automatically normalize newlines to the OS, so we just use an universal newline here
    changelog_file.write_text(f"{changelog_text}\n", encoding="utf-8")

    return str(changelog_file)


def write_changelog_files(
    runtime_ctx: RuntimeContext,
    release_history: ReleaseHistory,
    hvcs_client: HvcsBase,
    noop: bool = False,
) -> list[str]:
    project_dir = Path(runtime_ctx.repo_dir)
    template_dir = runtime_ctx.template_dir

    changelog_context = make_changelog_context(
        hvcs_client=hvcs_client,
        release_history=release_history,
        mode=runtime_ctx.changelog_mode,
        insertion_flag=runtime_ctx.changelog_insertion_flag,
        prev_changelog_file=runtime_ctx.changelog_file,
        mask_initial_release=runtime_ctx.changelog_mask_initial_release,
    )

    user_templates = []

    # Update known templates list if Directory exists and directory has actual files to render
    if template_dir.is_dir():
        user_templates.extend(
            [
                f
                for f in template_dir.rglob("*")
                if f.is_file() and f.suffix == JINJA2_EXTENSION
            ]
        )

        with suppress(ValueError):
            # do not include a release notes override when considering number of changelog templates
            user_templates.remove(template_dir / DEFAULT_RELEASE_NOTES_TPL_FILE)

    # Render user templates if found
    if len(user_templates) > 0:
        return apply_user_changelog_template_directory(
            template_dir=template_dir,
            environment=changelog_context.bind_to_environment(
                runtime_ctx.template_environment
            ),
            destination_dir=project_dir,
            noop=noop,
        )

    logger.info(
        "No contents found in %r, using default changelog template", template_dir
    )
    return [
        write_default_changelog(
            changelog_file=runtime_ctx.changelog_file,
            destination_dir=project_dir,
            output_format=runtime_ctx.changelog_output_format,
            changelog_context=changelog_context,
            changelog_style=runtime_ctx.changelog_style,
            noop=noop,
        )
    ]


def generate_release_notes(
    hvcs_client: HvcsBase,
    release: Release,
    template_dir: Path,
    history: ReleaseHistory,
    style: str,
    mask_initial_release: bool,
    license_name: str = "",
) -> str:
    users_tpl_file = template_dir / DEFAULT_RELEASE_NOTES_TPL_FILE

    # Determine if the user has a custom release notes template or we should use
    # the default template directory with our default release notes template
    tpl_dir = (
        template_dir
        if users_tpl_file.is_file()
        else get_default_tpl_dir(
            style=style, sub_dir=ChangelogOutputFormat.MARKDOWN.value
        )
    )

    release_notes_tpl_file = (
        users_tpl_file.name
        if users_tpl_file.is_file()
        else DEFAULT_RELEASE_NOTES_TPL_FILE
    )

    release_notes_env = ReleaseNotesContext(
        repo_name=hvcs_client.repo_name,
        repo_owner=hvcs_client.owner,
        hvcs_type=hvcs_client.__class__.__name__.lower(),
        version=release["version"],
        release=release,
        mask_initial_release=mask_initial_release,
        license_name=license_name,
        filters=(
            *hvcs_client.get_changelog_context_filters(),
            create_pypi_url,
            autofit_text_width,
            sort_numerically,
        ),
    ).bind_to_environment(
        # Use a new, non-configurable environment for release notes -
        # not user-configurable at the moment
        environment(autoescape=False, template_dir=tpl_dir)
    )

    # TODO: Remove in v11
    release_notes_env.globals["context"] = release_notes_env.globals["ctx"] = {
        "history": history,
        "mask_initial_release": mask_initial_release,
    }

    return render_release_notes(
        release_notes_template_file=release_notes_tpl_file,
        template_env=release_notes_env,
    )



import io
import logging
import random
import re
import string
from logging import LogRecord

import pytest

from semantic_release.cli.masking_filter import MaskingFilter

random.seed(0)


def _random_string(length: int = 10) -> str:
    alphabet = (
        string.ascii_lowercase
        + string.ascii_uppercase
        + string.digits
        + string.punctuation
    )
    return "".join(random.choice(alphabet) for _ in range(length))


@pytest.fixture
def default_masking_filter():
    return MaskingFilter()


@pytest.fixture
def logging_output_stream():
    return io.StringIO()


@pytest.fixture
def logger(logging_output_stream, default_masking_filter):
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.addHandler(logging.StreamHandler(logging_output_stream))
    log = logging.getLogger(__name__)
    log.setLevel(logging.DEBUG)

    for h in root.handlers:
        h.addFilter(default_masking_filter)

    return log


@pytest.mark.parametrize(
    "unwanted", [f(obj) for f in (repr, str) for obj in ("", None)]
)
def test_unwanted_masks_not_applied(default_masking_filter, unwanted):
    default_masking_filter.add_mask_for(unwanted, "foo")
    assert default_masking_filter._redact_patterns["foo"] == set()

    test_str = f"A long string containing the unwanted {unwanted} data"
    assert default_masking_filter.mask(test_str) == test_str


@pytest.mark.parametrize(
    "masked, secret",
    [
        ("secret-token", "secret-token"),
        (re.compile(r"ghp_.+?(?=\s|$)"), "ghp_" + _random_string(15)),
    ],
)
@pytest.mark.parametrize("use_named_masks", (True, False))
def test_mask_applied(use_named_masks, masked, secret):
    masker = MaskingFilter(_use_named_masks=use_named_masks)
    masker.add_mask_for(masked, "secret")
    test_str = "Your secret is... {secret} preferably hidden"

    assert masker.mask(test_str.format(secret=secret)) == test_str.format(
        secret="<'secret' (value removed)>" if use_named_masks else masker.REPLACE_STR
    )


_secrets = (
    "token" + _random_string(),
    "token" + _random_string(),
    "secret" + _random_string(),
    "secret" + _random_string(),
)


@pytest.mark.parametrize(
    "masked, secrets",
    [
        (_secrets, _secrets),
        ((re.compile(r"token.+?(?=\s|$)"), re.compile(r"secret.+?(?=\s|$)")), _secrets),
    ],
)
def test_multiple_secrets_with_same_mask(masked, secrets):
    masker = MaskingFilter(_use_named_masks=True)

    for mask in masked:
        masker.add_mask_for(mask, "ksam")

    test_str = " ".join(secrets)

    assert masker.mask(test_str) == " ".join(
        "<'ksam' (value removed)>" for _ in secrets
    )


def test_secrets_exact_replacement():
    masker = MaskingFilter(_use_named_masks=True)
    for secret in _secrets:
        masker.add_mask_for(secret, "smak")

    test_str = ", ".join(_secrets) + "!"
    assert (
        masker.mask(test_str)
        == ", ".join("<'smak' (value removed)>" for _ in _secrets) + "!"
    )


@pytest.mark.parametrize(
    "rec",
    [
        LogRecord(
            name=__name__,
            level=logging.INFO,
            pathname=__file__,
            lineno=10,
            args=(_secrets[3],),
            msg="long message with format %s for secret",
            exc_info=None,
        ),
        LogRecord(
            name=__name__,
            level=logging.INFO,
            pathname=__file__,
            lineno=10,
            args={"secret1": _secrets[1], "secret2": _secrets[2]},
            msg="another message using %(secret1)s and %(secret2)s",
            exc_info=None,
        ),
    ],
)
@pytest.mark.parametrize(
    "masked", (_secrets, (re.compile(r"(secret|token).+?(?=\s|$)"),))
)
def test_log_record_is_masked_with_simple_args(default_masking_filter, rec, masked):
    for mask in masked:
        default_masking_filter.add_mask_for(mask)

    if isinstance(rec.args, tuple):
        assert rec.msg % tuple(
            default_masking_filter.REPLACE_STR for _ in rec.args
        ) == default_masking_filter.mask(rec.getMessage())
    elif isinstance(rec.args, dict):
        assert rec.msg % {
            k: default_masking_filter.REPLACE_STR for k in rec.args
        } == default_masking_filter.mask(rec.getMessage())


@pytest.mark.parametrize(
    "rec",
    [
        LogRecord(
            name=__name__,
            level=logging.INFO,
            pathname=__file__,
            lineno=10,
            args=(_secrets,),
            msg="long message with format %s for secrets",
            exc_info=None,
        ),
        LogRecord(
            name=__name__,
            level=logging.INFO,
            pathname=__file__,
            lineno=10,
            args={"secret1": _secrets[1], "other": _secrets[2:]},
            msg="another message using %(secret1)s and %(other)r",
            exc_info=None,
        ),
    ],
)
@pytest.mark.parametrize(
    "masked", (_secrets, (re.compile(r"(secret|token).+?(?=\s|$)"),))
)
def test_log_record_is_masked_with_nontrivial_args(default_masking_filter, rec, masked):
    for mask in masked:
        default_masking_filter.add_mask_for(mask)

    assert any(secret in rec.getMessage() for secret in _secrets)
    assert all(
        secret not in default_masking_filter.mask(rec.getMessage())
        for secret in _secrets
    )


@pytest.mark.parametrize(
    "log_level",
    (
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ),
)
def test_log_messages_are_masked(
    default_masking_filter, log_level, logging_output_stream, logger, tmp_path
):
    for secret in _secrets:
        default_masking_filter.add_mask_for(secret)

    logger.log(log_level, ", ".join("%s" for _ in _secrets), *_secrets)

    root = logging.getLogger()
    for h in (*root.handlers, *logger.handlers):
        h.flush()

    written = logging_output_stream.getvalue()
    assert all(secret not in written for secret in _secrets)


@pytest.mark.parametrize("obj", (object(), (), {}, AttributeError("whoopsie")))
def test_non_strings_are_returned(default_masking_filter, obj):
    rec = LogRecord(
        name=__name__,
        level=logging.INFO,
        pathname=__file__,
        lineno=10,
        args=(),
        msg=obj,
        exc_info=None,
    )

    assert default_masking_filter.mask(rec.getMessage()) == str(obj)



"""Custom Errors"""


class SemanticReleaseBaseError(Exception):
    """
    Base Exception from which all other custom Exceptions defined in semantic_release
    inherit
    """


class InternalError(SemanticReleaseBaseError):
    """Raised when an internal error occurs, which should never happen"""


class InvalidConfiguration(SemanticReleaseBaseError):
    """Raised when configuration is deemed invalid"""


class InvalidParserOptions(InvalidConfiguration):
    """Raised when the parser options are invalid"""


class MissingGitRemote(SemanticReleaseBaseError):
    """Raised when repository is missing the configured remote origin or upstream"""


class InvalidVersion(ValueError, SemanticReleaseBaseError):
    """
    Raised when Version.parse attempts to parse a string containing
    an invalid version.
    """


class NotAReleaseBranch(InvalidConfiguration):
    """
    Raised when semantic_release is invoked on a branch which isn't configured for
    releases
    """


class DetachedHeadGitError(SemanticReleaseBaseError):
    """Raised when the git repository is in a detached HEAD state"""


class CommitParseError(SemanticReleaseBaseError):
    """
    Raised when a commit cannot be parsed by a commit parser. Custom commit parsers
    should also raise this Exception
    """


class MissingMergeBaseError(SemanticReleaseBaseError):
    """
    Raised when the merge base cannot be found with the current history. Generally
    because of a shallow git clone.
    """


class UnexpectedResponse(SemanticReleaseBaseError):
    """
    Raised when an HTTP response cannot be parsed properly or the expected structure
    is not found.
    """


class IncompleteReleaseError(SemanticReleaseBaseError):
    """
    Raised when there is a failure amongst one of the api requests when creating a
    release on a remote hvcs.
    """


class AssetUploadError(SemanticReleaseBaseError):
    """
    Raised when there is a failure uploading an asset to a remote hvcs's release artifact
    storage.
    """


class ParserLoadError(SemanticReleaseBaseError):
    """
    Raised when there is a failure to find, load, or instantiate a custom parser
    definition.
    """


class BuildDistributionsError(SemanticReleaseBaseError):
    """Raised when there is a failure to build the distribution files."""


class GitAddError(SemanticReleaseBaseError):
    """Raised when there is a failure to add files to the git index."""


class GitCommitError(SemanticReleaseBaseError):
    """Raised when there is a failure to commit the changes."""


class GitCommitEmptyIndexError(SemanticReleaseBaseError):
    """Raised when there is an attempt to commit an empty index."""


class GitTagError(SemanticReleaseBaseError):
    """Raised when there is a failure to tag the release."""


class GitPushError(SemanticReleaseBaseError):
    """Raised when there is a failure to push to the git remote."""


class GitFetchError(SemanticReleaseBaseError):
    """Raised when there is a failure to fetch from the git remote."""


class LocalGitError(SemanticReleaseBaseError):
    """Raised when there is a failure with local git operations."""


class UnknownUpstreamBranchError(SemanticReleaseBaseError):
    """Raised when the upstream branch cannot be determined."""


class UpstreamBranchChangedError(SemanticReleaseBaseError):
    """Raised when the upstream branch has changed before pushing."""



from __future__ import annotations

from pathlib import Path

import pytest


@pytest.hookimpl(tryfirst=True)
def pytest_collection_modifyitems(items: list[pytest.Item]) -> None:
    """Apply the unit marker to all tests in the unit test directory."""
    unit_test_directory = Path(__file__).parent
    for item in items:
        if unit_test_directory in item.path.parents:
            item.add_marker(pytest.mark.unit)
            if "order" not in [mark.name for mark in item.own_markers]:
                item.add_marker(pytest.mark.order("first"))



from tests.fixtures.repos.git_flow.repo_w_1_release_channel import *
from tests.fixtures.repos.git_flow.repo_w_2_release_channels import *
from tests.fixtures.repos.git_flow.repo_w_3_release_channels import *
from tests.fixtures.repos.git_flow.repo_w_4_release_channels import *



from __future__ import annotations

from collections import defaultdict
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, TypedDict

from git.objects.tag import TagObject

from semantic_release.commit_parser import ParseError
from semantic_release.commit_parser.token import ParsedCommit
from semantic_release.commit_parser.util import force_str
from semantic_release.enums import LevelBump
from semantic_release.globals import logger
from semantic_release.helpers import validate_types_in_sequence
from semantic_release.version.algorithm import tags_and_versions

if TYPE_CHECKING:  # pragma: no cover
    from re import Pattern
    from typing import Iterable, Iterator

    from git.repo.base import Repo
    from git.util import Actor

    from semantic_release.commit_parser import (
        CommitParser,
        ParseResult,
        ParserOptions,
    )
    from semantic_release.version.translator import VersionTranslator
    from semantic_release.version.version import Version


class ReleaseHistory:
    @classmethod
    def from_git_history(
        cls,
        repo: Repo,
        translator: VersionTranslator,
        commit_parser: CommitParser[ParseResult, ParserOptions],
        exclude_commit_patterns: Iterable[Pattern[str]] = (),
    ) -> ReleaseHistory:
        all_git_tags_and_versions = tags_and_versions(repo.tags, translator)
        unreleased: dict[str, list[ParseResult]] = defaultdict(list)
        released: dict[Version, Release] = {}

        # Performance optimization: create a mapping of tag sha to version
        # so we can quickly look up the version for a given commit based on sha
        tag_sha_2_version_lookup = {
            tag.commit.hexsha: (tag, version)
            for tag, version in all_git_tags_and_versions
        }

        ignore_merge_commits = bool(
            hasattr(commit_parser, "options")
            and hasattr(commit_parser.options, "ignore_merge_commits")
            and getattr(commit_parser.options, "ignore_merge_commits")  # noqa: B009
        )

        # Strategy:
        # Loop through commits in history, parsing as we go.
        # Add these commits to `unreleased` as a key-value mapping
        # of type_ to ParseResult, until we encounter a tag
        # which matches a commit.
        # Then, we add the version for that tag as a key to `released`,
        # and set the value to an empty dict. Into that empty dict
        # we place the key-value mapping type_ to ParseResult as before.
        # We do this until we encounter a commit which another tag matches.

        the_version: Version | None = None

        for commit in repo.iter_commits("HEAD", topo_order=True):
            # Determine if we have found another release
            logger.debug("checking if commit %s matches any tags", commit.hexsha[:7])
            t_v = tag_sha_2_version_lookup.get(commit.hexsha, None)

            if t_v is None:
                logger.debug("no tags correspond to commit %s", commit.hexsha)
            else:
                # Unpack the tuple (overriding the current version)
                tag, the_version = t_v
                # we have found the latest commit introduced by this tag
                # so we create a new Release entry
                logger.debug("found commit %s for tag %s", commit.hexsha, tag.name)

                # tag.object is a Commit if the tag is lightweight, otherwise
                # it is a TagObject with additional metadata about the tag
                if isinstance(tag.object, TagObject):
                    tagger = tag.object.tagger
                    committer = tag.object.tagger.committer()
                    _tz = timezone(timedelta(seconds=-1 * tag.object.tagger_tz_offset))
                    tagged_date = datetime.fromtimestamp(tag.object.tagged_date, tz=_tz)
                else:
                    # For some reason, sometimes tag.object is a Commit
                    tagger = tag.object.author
                    committer = tag.object.author
                    _tz = timezone(timedelta(seconds=-1 * tag.object.author_tz_offset))
                    tagged_date = datetime.fromtimestamp(
                        tag.object.committed_date, tz=_tz
                    )

                release = Release(
                    tagger=tagger,
                    committer=committer,
                    tagged_date=tagged_date,
                    elements=defaultdict(list),
                    version=the_version,
                )

                released.setdefault(the_version, release)

            logger.info(
                "parsing commit [%s] %s",
                commit.hexsha[:8],
                str(commit.message).replace("\n", " ")[:54],
            )
            # returns a ParseResult or list of ParseResult objects,
            # it is usually one, but we split a commit if a squashed merge is detected
            parse_results = commit_parser.parse(commit)

            if not any(
                (
                    isinstance(parse_results, (ParseError, ParsedCommit)),
                    (
                        (
                            isinstance(parse_results, list)
                            or type(parse_results) == tuple
                        )
                        and validate_types_in_sequence(
                            parse_results, (ParseError, ParsedCommit)
                        )
                    ),
                )
            ):
                raise TypeError("Unexpected type returned from commit_parser.parse")

            results: list[ParseResult] = [
                *(
                    [parse_results]
                    if isinstance(parse_results, (ParseError, ParsedCommit))
                    else parse_results
                ),
            ]

            is_squash_commit = bool(len(results) > 1)

            # iterate through parsed commits to add to changelog definition
            for parsed_result in results:
                commit_message = force_str(parsed_result.commit.message)
                commit_type = (
                    "unknown"
                    if isinstance(parsed_result, ParseError)
                    else parsed_result.type
                )
                logger.debug("commit has type '%s'", commit_type)

                has_exclusion_match = any(
                    pattern.match(commit_message) for pattern in exclude_commit_patterns
                )

                commit_level_bump = (
                    LevelBump.NO_RELEASE
                    if isinstance(parsed_result, ParseError)
                    else parsed_result.bump
                )

                if ignore_merge_commits and parsed_result.is_merge_commit():
                    logger.info("Excluding merge commit[%s]", parsed_result.short_hash)
                    continue

                # Skip excluded commits except for any commit causing a version bump
                # Reasoning: if a commit causes a version bump, and no other commits
                # are included, then the changelog will be empty. Even if ther was other
                # commits included, the true reason for a version bump would be missing.
                if has_exclusion_match and commit_level_bump == LevelBump.NO_RELEASE:
                    logger.info(
                        "Excluding %s commit[%s] %s",
                        "piece of squashed" if is_squash_commit else "",
                        parsed_result.short_hash,
                        commit_message.split("\n", maxsplit=1)[0][:20],
                    )
                    continue

                if (
                    isinstance(parsed_result, ParsedCommit)
                    and not parsed_result.include_in_changelog
                ):
                    logger.info(
                        str.join(
                            " ",
                            [
                                "Excluding commit[%s] because parser determined",
                                "it should not included in the changelog",
                            ],
                        ),
                        parsed_result.short_hash,
                    )
                    continue

                if the_version is None:
                    logger.info(
                        "[Unreleased] adding commit[%s] to unreleased '%s'",
                        parsed_result.short_hash,
                        commit_type,
                    )
                    unreleased[commit_type].append(parsed_result)
                    continue

                logger.info(
                    "[%s] adding commit[%s] to release '%s'",
                    the_version,
                    parsed_result.short_hash,
                    commit_type,
                )

                released[the_version]["elements"][commit_type].append(parsed_result)

        return cls(unreleased=unreleased, released=released)

    def __init__(
        self, unreleased: dict[str, list[ParseResult]], released: dict[Version, Release]
    ) -> None:
        self.released = released
        self.unreleased = unreleased

    def __iter__(
        self,
    ) -> Iterator[dict[str, list[ParseResult]] | dict[Version, Release]]:
        """
        Enables unpacking:

        >>> rh = ReleaseHistory(...)
        >>> unreleased, released = rh
        """
        yield self.unreleased
        yield self.released

    def release(
        self, version: Version, tagger: Actor, committer: Actor, tagged_date: datetime
    ) -> ReleaseHistory:
        if version in self.released:
            raise ValueError(f"{version} has already been released!")

        # return a new instance to avoid potential accidental
        # mutation
        return ReleaseHistory(
            unreleased={},
            released={
                version: {
                    "tagger": tagger,
                    "committer": committer,
                    "tagged_date": tagged_date,
                    "elements": self.unreleased,
                    "version": version,
                },
                **self.released,
            },
        )

    def __repr__(self) -> str:
        return (
            f"<{type(self).__qualname__}: "
            f"{sum(len(commits) for commits in self.unreleased.values())} "
            f"commits unreleased, {len(self.released)} versions released>"
        )


class Release(TypedDict):
    tagger: Actor
    committer: Actor
    tagged_date: datetime
    elements: dict[str, list[ParseResult]]
    version: Version



from __future__ import annotations

from datetime import timedelta
from itertools import count
from pathlib import Path
from typing import TYPE_CHECKING, cast

import pytest

from semantic_release.cli.config import ChangelogOutputFormat

import tests.conftest
import tests.const
import tests.util
from tests.const import (
    EXAMPLE_HVCS_DOMAIN,
    INITIAL_COMMIT_MESSAGE,
    RepoActionStep,
)

if TYPE_CHECKING:
    from typing import Any, Sequence

    from semantic_release.commit_parser._base import CommitParser, ParserOptions
    from semantic_release.commit_parser.conventional.parser import (
        ConventionalCommitParser,
    )
    from semantic_release.commit_parser.emoji import EmojiCommitParser
    from semantic_release.commit_parser.scipy import ScipyCommitParser
    from semantic_release.commit_parser.token import ParseResult

    from tests.conftest import (
        GetCachedRepoDataFn,
        GetMd5ForSetOfFilesFn,
        GetStableDateNowFn,
    )
    from tests.fixtures.example_project import ExProjectDir
    from tests.fixtures.git_repo import (
        BuildRepoFromDefinitionFn,
        BuildRepoOrCopyCacheFn,
        BuildSpecificRepoFn,
        BuiltRepoResult,
        CommitConvention,
        ConvertCommitSpecsToCommitDefsFn,
        ExProjectGitRepoFn,
        GetRepoDefinitionFn,
        RepoActions,
        TomlSerializableTypes,
    )


@pytest.fixture(scope="session")
def deps_files_4_repo_w_no_tags(
    deps_files_4_example_git_project: list[Path],
) -> list[Path]:
    return [
        *deps_files_4_example_git_project,
        # This file
        Path(__file__).absolute(),
        # because of imports
        Path(tests.const.__file__).absolute(),
        Path(tests.util.__file__).absolute(),
        # because of the fixtures
        Path(tests.conftest.__file__).absolute(),
    ]


@pytest.fixture(scope="session")
def build_spec_hash_4_repo_w_no_tags(
    get_md5_for_set_of_files: GetMd5ForSetOfFilesFn,
    deps_files_4_repo_w_no_tags: list[Path],
) -> str:
    # Generates a hash of the build spec to set when to invalidate the cache
    return get_md5_for_set_of_files(deps_files_4_repo_w_no_tags)


@pytest.fixture(scope="session")
def get_repo_definition_4_trunk_only_repo_w_no_tags(
    convert_commit_specs_to_commit_defs: ConvertCommitSpecsToCommitDefsFn,
    changelog_md_file: Path,
    changelog_rst_file: Path,
    stable_now_date: GetStableDateNowFn,
    default_conventional_parser: ConventionalCommitParser,
    default_emoji_parser: EmojiCommitParser,
    default_scipy_parser: ScipyCommitParser,
) -> GetRepoDefinitionFn:
    """
    Builds a repository with trunk-only committing (no-branching) strategy without
    any releases.
    """
    parser_classes: dict[CommitConvention, CommitParser[Any, Any]] = {
        "conventional": default_conventional_parser,
        "emoji": default_emoji_parser,
        "scipy": default_scipy_parser,
    }

    def _get_repo_from_definition(
        commit_type: CommitConvention,
        hvcs_client_name: str = "github",
        hvcs_domain: str = EXAMPLE_HVCS_DOMAIN,
        tag_format_str: str | None = None,
        extra_configs: dict[str, TomlSerializableTypes] | None = None,
        mask_initial_release: bool = True,
        ignore_merge_commits: bool = True,
    ) -> Sequence[RepoActions]:
        stable_now_datetime = stable_now_date()
        commit_timestamp_gen = (
            (stable_now_datetime + timedelta(seconds=i)).isoformat(timespec="seconds")
            for i in count(step=1)
        )

        repo_construction_steps: list[RepoActions] = []
        repo_construction_steps.extend(
            [
                {
                    "action": RepoActionStep.CONFIGURE,
                    "details": {
                        "commit_type": commit_type,
                        "hvcs_client_name": hvcs_client_name,
                        "hvcs_domain": hvcs_domain,
                        "tag_format_str": tag_format_str,
                        "mask_initial_release": mask_initial_release,
                        "extra_configs": {
                            # Set the default release branch
                            "tool.semantic_release.branches.main": {
                                "match": r"^(main|master)$",
                                "prerelease": False,
                            },
                            "tool.semantic_release.commit_parser_options.ignore_merge_commits": ignore_merge_commits,
                            **(extra_configs or {}),
                        },
                    },
                },
                {
                    "action": RepoActionStep.MAKE_COMMITS,
                    "details": {
                        "commits": convert_commit_specs_to_commit_defs(
                            [
                                {
                                    "cid": (cid_c1_initial := "c1_initial_commit"),
                                    "conventional": INITIAL_COMMIT_MESSAGE,
                                    "emoji": INITIAL_COMMIT_MESSAGE,
                                    "scipy": INITIAL_COMMIT_MESSAGE,
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": bool(
                                        commit_type == "emoji"
                                    ),
                                },
                                {
                                    "cid": (cid_c2_feat1 := "c2-feat1"),
                                    "conventional": "feat: add new feature",
                                    "emoji": ":sparkles: add new feature",
                                    "scipy": "ENH: add new feature",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                                {
                                    "cid": (cid_c3_fix1 := "c3-fix1"),
                                    "conventional": "fix: correct some text",
                                    "emoji": ":bug: correct some text",
                                    "scipy": "MAINT: correct some text",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                                {
                                    "cid": (cid_c4_fix2 := "c4-fix2"),
                                    "conventional": "fix: correct more text\n\nCloses: #123",
                                    "emoji": ":bug: correct more text\n\nCloses: #123",
                                    "scipy": "MAINT: correct more text\n\nCloses: #123",
                                    "datetime": next(commit_timestamp_gen),
                                    "include_in_changelog": True,
                                },
                            ],
                            commit_type,
                            parser=cast(
                                "CommitParser[ParseResult, ParserOptions]",
                                parser_classes[commit_type],
                            ),
                        ),
                    },
                },
                {
                    "action": RepoActionStep.WRITE_CHANGELOGS,
                    "details": {
                        "new_version": "Unreleased",
                        "dest_files": [
                            {
                                "path": changelog_md_file,
                                "format": ChangelogOutputFormat.MARKDOWN,
                                "mask_initial_release": mask_initial_release,
                            },
                            {
                                "path": changelog_rst_file,
                                "format": ChangelogOutputFormat.RESTRUCTURED_TEXT,
                                "mask_initial_release": mask_initial_release,
                            },
                        ],
                        "commit_ids": [
                            cid_c1_initial,
                            cid_c2_feat1,
                            cid_c3_fix1,
                            cid_c4_fix2,
                        ],
                    },
                },
            ]
        )

        return repo_construction_steps

    return _get_repo_from_definition


@pytest.fixture(scope="session")
def build_trunk_only_repo_w_no_tags(
    build_repo_from_definition: BuildRepoFromDefinitionFn,
    get_repo_definition_4_trunk_only_repo_w_no_tags: GetRepoDefinitionFn,
    get_cached_repo_data: GetCachedRepoDataFn,
    build_repo_or_copy_cache: BuildRepoOrCopyCacheFn,
    build_spec_hash_4_repo_w_no_tags: str,
) -> BuildSpecificRepoFn:
    def _build_specific_repo_type(
        repo_name: str, commit_type: CommitConvention, dest_dir: Path
    ) -> Sequence[RepoActions]:
        def _build_repo(cached_repo_path: Path) -> Sequence[RepoActions]:
            repo_construction_steps = get_repo_definition_4_trunk_only_repo_w_no_tags(
                commit_type=commit_type,
            )
            return build_repo_from_definition(cached_repo_path, repo_construction_steps)

        build_repo_or_copy_cache(
            repo_name=repo_name,
            build_spec_hash=build_spec_hash_4_repo_w_no_tags,
            build_repo_func=_build_repo,
            dest_dir=dest_dir,
        )

        if not (cached_repo_data := get_cached_repo_data(proj_dirname=repo_name)):
            raise ValueError("Failed to retrieve repo data from cache")

        return cached_repo_data["build_definition"]

    return _build_specific_repo_type


# --------------------------------------------------------------------------- #
# Test-level fixtures that will cache the built directory & set up test case  #
# --------------------------------------------------------------------------- #


@pytest.fixture
def repo_w_no_tags_conventional_commits_using_tag_format(
    build_repo_from_definition: BuildRepoFromDefinitionFn,
    get_repo_definition_4_trunk_only_repo_w_no_tags: GetRepoDefinitionFn,
    get_cached_repo_data: GetCachedRepoDataFn,
    build_repo_or_copy_cache: BuildRepoOrCopyCacheFn,
    build_spec_hash_4_repo_w_no_tags: str,
    example_project_git_repo: ExProjectGitRepoFn,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
) -> BuiltRepoResult:
    """
    Replicates repo with no tags, but with a tag format X{version}

    Follows tag format defined in python-semantic-release#1137
    """
    repo_name = repo_w_no_tags_conventional_commits_using_tag_format.__name__
    commit_type: CommitConvention = (
        repo_name.split("_commits", maxsplit=1)[0].split("_")[-1]  # type: ignore[assignment]
    )

    def _build_repo(cached_repo_path: Path) -> Sequence[RepoActions]:
        repo_construction_steps = get_repo_definition_4_trunk_only_repo_w_no_tags(
            commit_type=commit_type,
            tag_format_str="X{version}",
        )
        return build_repo_from_definition(cached_repo_path, repo_construction_steps)

    build_repo_or_copy_cache(
        repo_name=repo_name,
        build_spec_hash=build_spec_hash_4_repo_w_no_tags,
        build_repo_func=_build_repo,
        dest_dir=example_project_dir,
    )

    if not (cached_repo_data := get_cached_repo_data(proj_dirname=repo_name)):
        raise ValueError("Failed to retrieve repo data from cache")

    return {
        "definition": cached_repo_data["build_definition"],
        "repo": example_project_git_repo(),
    }


@pytest.fixture
def repo_w_no_tags_conventional_commits_w_zero_version(
    build_repo_from_definition: BuildRepoFromDefinitionFn,
    get_repo_definition_4_trunk_only_repo_w_no_tags: GetRepoDefinitionFn,
    get_cached_repo_data: GetCachedRepoDataFn,
    build_repo_or_copy_cache: BuildRepoOrCopyCacheFn,
    build_spec_hash_4_repo_w_no_tags: str,
    example_project_git_repo: ExProjectGitRepoFn,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
) -> BuiltRepoResult:
    """Replicates repo with no tags, but with allow_zero_version=True"""
    repo_name = repo_w_no_tags_conventional_commits_w_zero_version.__name__
    commit_type: CommitConvention = (
        repo_name.split("_commits", maxsplit=1)[0].split("_")[-1]  # type: ignore[assignment]
    )

    def _build_repo(cached_repo_path: Path) -> Sequence[RepoActions]:
        repo_construction_steps = get_repo_definition_4_trunk_only_repo_w_no_tags(
            commit_type=commit_type,
            extra_configs={
                "tool.semantic_release.allow_zero_version": True,
            },
        )
        return build_repo_from_definition(cached_repo_path, repo_construction_steps)

    build_repo_or_copy_cache(
        repo_name=repo_name,
        build_spec_hash=build_spec_hash_4_repo_w_no_tags,
        build_repo_func=_build_repo,
        dest_dir=example_project_dir,
    )

    if not (cached_repo_data := get_cached_repo_data(proj_dirname=repo_name)):
        raise ValueError("Failed to retrieve repo data from cache")

    return {
        "definition": cached_repo_data["build_definition"],
        "repo": example_project_git_repo(),
    }


@pytest.fixture
def repo_w_no_tags_conventional_commits_unmasked_initial_release(
    build_repo_from_definition: BuildRepoFromDefinitionFn,
    get_repo_definition_4_trunk_only_repo_w_no_tags: GetRepoDefinitionFn,
    get_cached_repo_data: GetCachedRepoDataFn,
    build_repo_or_copy_cache: BuildRepoOrCopyCacheFn,
    build_spec_hash_4_repo_w_no_tags: str,
    example_project_git_repo: ExProjectGitRepoFn,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
) -> BuiltRepoResult:
    """Replicates repo with no tags, but with mask_initial_release=False"""
    repo_name = repo_w_no_tags_conventional_commits_unmasked_initial_release.__name__
    commit_type: CommitConvention = (
        repo_name.split("_commits", maxsplit=1)[0].split("_")[-1]  # type: ignore[assignment]
    )

    def _build_repo(cached_repo_path: Path) -> Sequence[RepoActions]:
        repo_construction_steps = get_repo_definition_4_trunk_only_repo_w_no_tags(
            commit_type=commit_type,
            extra_configs={
                "tool.semantic_release.changelog.default_templates.mask_initial_release": False,
            },
        )
        return build_repo_from_definition(cached_repo_path, repo_construction_steps)

    build_repo_or_copy_cache(
        repo_name=repo_name,
        build_spec_hash=build_spec_hash_4_repo_w_no_tags,
        build_repo_func=_build_repo,
        dest_dir=example_project_dir,
    )

    if not (cached_repo_data := get_cached_repo_data(proj_dirname=repo_name)):
        raise ValueError("Failed to retrieve repo data from cache")

    return {
        "definition": cached_repo_data["build_definition"],
        "repo": example_project_git_repo(),
    }


@pytest.fixture
def repo_w_no_tags_conventional_commits(
    build_trunk_only_repo_w_no_tags: BuildSpecificRepoFn,
    example_project_git_repo: ExProjectGitRepoFn,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
) -> BuiltRepoResult:
    repo_name = repo_w_no_tags_conventional_commits.__name__
    commit_type: CommitConvention = repo_name.split("_")[-2]  # type: ignore[assignment]

    return {
        "definition": build_trunk_only_repo_w_no_tags(
            repo_name=repo_name,
            commit_type=commit_type,
            dest_dir=example_project_dir,
        ),
        "repo": example_project_git_repo(),
    }


@pytest.fixture
def repo_w_no_tags_emoji_commits(
    build_trunk_only_repo_w_no_tags: BuildSpecificRepoFn,
    example_project_git_repo: ExProjectGitRepoFn,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
) -> BuiltRepoResult:
    repo_name = repo_w_no_tags_emoji_commits.__name__
    commit_type: CommitConvention = repo_name.split("_")[-2]  # type: ignore[assignment]

    return {
        "definition": build_trunk_only_repo_w_no_tags(
            repo_name=repo_name,
            commit_type=commit_type,
            dest_dir=example_project_dir,
        ),
        "repo": example_project_git_repo(),
    }


@pytest.fixture
def repo_w_no_tags_scipy_commits(
    build_trunk_only_repo_w_no_tags: BuildSpecificRepoFn,
    example_project_git_repo: ExProjectGitRepoFn,
    example_project_dir: ExProjectDir,
    change_to_ex_proj_dir: None,
) -> BuiltRepoResult:
    repo_name = repo_w_no_tags_scipy_commits.__name__
    commit_type: CommitConvention = repo_name.split("_")[-2]  # type: ignore[assignment]

    return {
        "definition": build_trunk_only_repo_w_no_tags(
            repo_name=repo_name,
            commit_type=commit_type,
            dest_dir=example_project_dir,
        ),
        "repo": example_project_git_repo(),
    }
